# pylint: disable=line-too-long
"""
Will store all the functions to play with and modify the animations and
waypoints in Synfig format
"""

import sys
import copy
from lxml import etree
import settings
from common.misc import approximate_equal, is_animated, get_frame
from common.Vector import Vector
from helpers.bezier import get_bezier_val
sys.path.append("..")


def gen_dummy_waypoint(non_animated, animated_tag, anim_type, animated_name="anything"):
	"""
	Makes a non animated parameter to animated parameter by creating a new dummy
	waypoint with constant animation

	Args:
		non_animated (lxml.etree._Element): Holds the non-animated parameter in Synfig xml format
		animated_tag(str)                : Decides the tag of the animation
		anim_type    (str)                : Decides the animation type

	Returns:
		(lxml.etree._Element) : Updated non-animated parameter, which is now animated
	"""
	is_animate = is_animated(non_animated[0])
	if is_animate == settings.ANIMATED:
		# If already animated, no need to add waypoints
		# Forcibly set it's animation type to the given anim_type :needed in:->
		# properties/shapePropKeyframe.py #31
		non_animated[0].attrib["type"] = anim_type
		return non_animated
	elif is_animate == settings.NOT_ANIMATED:
		st = '<{animated_tag} name="{animated_name}"><animated type="{anim_type}"><waypoint time="0s" before="constant" after="constant"></waypoint></animated></{animated_tag}>'
		st = st.format(anim_type=anim_type, animated_name=animated_name, animated_tag=animated_tag)
		root = etree.fromstring(st)
		root[0][0].append(copy.deepcopy(non_animated[0]))
		non_animated = root
	elif is_animate == settings.SINGLE_WAYPOINT:
		# Forcibly set it's animation type to the given anim_type
		non_animated[0].attrib["type"] = anim_type

		non_animated[0][0].attrib["before"] = non_animated[0][0].attrib["after"] = "constant"

	new_waypoint = copy.deepcopy(non_animated[0][0])
	frame = get_frame(non_animated[0][0])
	frame += 1
	time = frame / settings.lottie_format["fr"]
	time = str(time) + "s"
	new_waypoint.attrib["time"] = time
	non_animated[0].insert(1, new_waypoint)
	return non_animated


def insert_waypoint_at_frame(animated, orig_path, frame, animated_name):
	"""
	This function will only insert a waypoint at 'frame' if no waypoint is
	present at that 'frame' already

	Args:
		animated      (lxml.etree._Element): Holds the animation in Synfig xml format
		orig_path     (dict)               : Holds the animation in Lottie format
		frame         (int)                : The frame at which the waypoint is to be inserted
		animated_name (str)                : The name/type of animation

	Returns:
		(None)
	"""
	i = 0
	while i < len(animated):
		at_frame = get_frame(animated[i])
		if approximate_equal(frame, at_frame):
			return
		elif frame < at_frame:
			break
		i += 1
	pos = get_vector_at_frame(orig_path, frame)
	pos = to_Synfig_axis(pos, animated_name)

	if i == len(animated):
		new_waypoint = copy.deepcopy(animated[i-1])
	else:
		new_waypoint = copy.deepcopy(animated[i])
	if animated_name == "vector":
		new_waypoint[0][0].text = str(pos[0])
		new_waypoint[0][1].text = str(pos[1])
	else:
		new_waypoint[0].attrib["value"] = str(pos)

	new_waypoint.attrib["time"] = str(frame/settings.lottie_format["fr"]) + "s"
	if i == 0 or i == len(animated):
		# No need of tcb value copy as halt interpolation need to be copied here
		new_waypoint.attrib["before"] = new_waypoint.attrib["after"] = "constant"
	else:
		copy_tcb_average(new_waypoint, animated[i], animated[i-1])
		new_waypoint.attrib["before"] = animated[i-1].attrib["after"]
		new_waypoint.attrib["after"] = animated[i].attrib["before"]
		# If the interval was constant before, then the whole interval should
		# remain constant now also
		if new_waypoint.attrib["before"] == "constant" or new_waypoint.attrib["after"] == "constant":
			new_waypoint.attrib["before"] = new_waypoint.attrib["after"] = "constant"
	animated.insert(i, new_waypoint)


def print_animation(*argv):
	"""
	Given any animation, b, It prints the animation in pretty way.
	Helpful in debugging

	Args:
		argv (tuple): Holds the animations to be printed
		arg will be of type lxml.etree._Element

	Returns:
		(None)
	"""
	print("####################### START ########################")
	for arg in argv:
		a = copy.deepcopy(arg)
		print(etree.tostring(a, method='xml', encoding='utf8', pretty_print=True).decode())
	print("####################### END ##########################")


def to_Lottie_axis(pos, animated_name):
    """
    Opposite effecto of 'to_Synfig_axis'
    """
    if animated_name == "vector":
        pos[0] *= settings.PIX_PER_UNIT
        pos[1] *= settings.PIX_PER_UNIT
        pos[1] = -pos[1]
        pos[0], pos[1] = pos[0] + settings.lottie_format["w"]/2, pos[1] + settings.lottie_format["h"]/2
        return pos
    # should never reach here
    assert(False)


def to_Synfig_axis(pos, animated_name):
	"""
	Converts a Lottie format vector or values into Synfig format vector or
	values

	Args:
		pos             (float | list) : Stores the position/value to be converted to the Synfig format
		animated_name   (str)          : Distinguishes between position and value

	Returns:
		(float) If pos is of type float
		(list)  Else
	"""
	if animated_name == "vector":
		pos[0], pos[1] = pos[0] - settings.lottie_format["w"]/2, pos[1] - settings.lottie_format["h"]/2
		pos[1] = -pos[1]
		ret = [x/settings.PIX_PER_UNIT for x in pos]
	elif animated_name == "real":
		ret = pos / settings.PIX_PER_UNIT
	elif animated_name in ["amount", "opacity"]:
		ret = pos / settings.OPACITY_CONSTANT
	elif animated_name == "effects_opacity":
		ret = pos
	elif animated_name == "angle":
		s1 = int(pos / 360)
		t1 = pos % 360
		t1 = (90 - t1) % 360
		t1 = t1 + s1 * 360
		ret = t1
	return ret


def get_first_control_point(interval):
	"""
	Returns the first control point of a bezier interval

	Args:
		interval (dict) : Holds one interval of the bezier curve that is two waypoints

	Returns:
		(common.Vector.Vector) If the interval holds position bezier
		(float)       Else : the interval holds value bezier
	"""
	if len(interval["s"]) >= 2:
		st = Vector(interval["s"][0], interval["s"][1])
	else:
		st = interval["s"][0]
	return st


def get_last_control_point(interval):
	"""
	Returns the last control point of a bezier interval

	Args:
		interval (dict) : Holds one interval of the bezier curve that is two waypoints

	Returns:
		(common.Vector.Vector) If the interval holds position bezier
		(float)       Else : the interval holds value bezier
	"""
	if len(interval["e"]) >= 2:
		en = Vector(interval["e"][0], interval["e"][1])
	else:
		en = interval["e"][0]
	return en


def get_control_points(interval):
	"""
	Returns all 4 control points of a bezier interval

	Args:
		interval (dict) : Holds one interval of the bezier curve that is two waypoints

	Returns:
		(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector, common.Vector.Vector) If the interval holds position bezier
		(float, float, float, float)       Else : the interval holds value bezier
	"""
	# If the interval is for position or vector
	if "to" in interval.keys():
		st = Vector(interval["s"][0], interval["s"][1])
		en = Vector(interval["e"][0], interval["e"][1])
		to = Vector(interval["synfig_to"][0], interval["synfig_to"][1])
		ti = Vector(interval["synfig_ti"][0], interval["synfig_ti"][1])

	# If the interval is for real values
	else:
		st = interval["s"][0]
		en = interval["e"][0]
		to = interval["synfig_o"][0]
		ti = interval["synfig_i"][0]
	return st, to, ti, en


def get_vector_at_frame(path, t):
	"""
	Given 'path' in lottie format and t(in frames), this function returns the
	vector or real value at frame t depending on the type of path supplied to it

	Args:
		path (dict): Contains the bezier curve in Lottie JSON format
		t    (int) : Contains the frame at which the vector/value is requested

	Returns:
		(list)  If a positional bezier is queried
		(float) If a value bezier is queried
	"""
	keyfr = path["k"]
	# Find the interval in which it lies
	i = 0
	while i < len(keyfr):
		t_key = keyfr[i]["t"]
		if t < t_key:
			break
		i += 1
	i -= 1
	if i < 0:
		pos = get_first_control_point(keyfr[0])
	elif i < len(keyfr) - 1:
		# If hold interpolation
		if 'h' in keyfr[i].keys():
			pos = get_first_control_point(keyfr[i])
		else:
			st, to, ti, en = get_control_points(keyfr[i])
			this_fr = keyfr[i]["t"]
			next_fr = keyfr[i+1]["t"]
			percent = (t - this_fr) / (next_fr - this_fr)
			pos = get_bezier_val(st, st + to, en - ti, en, percent)
	elif i >= len(keyfr) - 1:
		pos = get_last_control_point(keyfr[-2])

	if isinstance(pos, Vector):
		return pos.get_list()
	return pos


def get_bool_at_frame(anim, frame):
	"""
	Calculates the boolean value at a given frame, given a boolean animation

	Args:
		anim  (lxml.etree._Element): Boolean animation
		frame (int)                : Frame at which the value is to be calculated

	Returns:
		(bool) : True if the value is "true" at that frame
			   : False otherwise
	"""
	i = 0
	not_convert_flag = False
	while i < len(anim):
		cur_fr = get_frame(anim[i])
		if frame <= cur_fr:
			break
		i += 1
	if i == 0:
		if anim[0][0].tag == 'bool':
			val = anim[0][0].attrib["value"]
		else:
			not_convert_flag = True
			val = anim[0][0][0][0][0][0].attrib["value"]
	elif i < len(anim):
		# frame lies between i-1 and i'th value of animation
		if anim[i-1][0].tag == 'bool':
			prev = anim[i-1][0].attrib["value"]
		else:
			prev = anim[i-1][0][0][0][0][0].attrib["value"]

		if anim[i][0].tag == 'bool':
			cur = anim[i][0].attrib["value"]
		else:
			cur = anim[i][0][0][0][0][0].attrib["value"]

		cur_frame = get_frame(anim[i])
		if frame == cur_frame:
			val = cur
		else:
			if prev == "true" and cur == "false":
				val = prev
			elif prev == "false" and cur == "true":
				val = cur
			elif prev == "true" and cur == "true":
				val = cur
			elif prev == "false" and cur == "false":
				val = cur
	elif i >= len(anim):
		if anim[-1][0].tag == 'bool':
			val = anim[-1][0].attrib["value"]
		else:
			val = anim[-1][0][0][0][0][0].attrib["value"]

	if val == "false" and not not_convert_flag:
		val = False
	elif val == "true" and not not_convert_flag:
		val = True
	elif val == "false" and not_convert_flag:
		val = True
	elif val == "true" and not_convert_flag:
		val = False
	return val


def waypoint_at_frame(anim, frame):
	"""
	Returns true if a waypoint is present at 'frame'
	"""
	for waypoint in anim:
		fr = get_frame(waypoint)
		if fr == frame:
			return True
	return False

def modify_bool_animation(anim):
	"""
	Inserts waypoints at such frames so that the animation is similar to that in
	lottie
	"""
	i = 0
	while i < len(anim):
		cur_fr = get_frame(anim[i])
		val_now = get_bool_at_frame(anim, cur_fr)
		# Check at one frame less and one frame more: only cases
		if cur_fr - 1 >= settings.lottie_format["ip"] and not waypoint_at_frame(anim, cur_fr-1):
			val_before = get_bool_at_frame(anim, cur_fr - 1)
			if val_now != val_before:
				new = copy.deepcopy(anim[i])
				new.attrib["time"] = str((cur_fr - 1)/settings.lottie_format["fr"]) + "s"
				if val_before:
					new[0].attrib["value"] = "true"
				else:
					new[0].attrib["value"] = "false"
				anim.insert(i, new)
				i += 1
		if cur_fr + 1 <= settings.lottie_format["op"] and not waypoint_at_frame(anim, cur_fr + 1):
			val_after = get_bool_at_frame(anim, cur_fr + 1)
			if val_after != val_now:
				new = copy.deepcopy(anim[i])
				new.attrib["time"] = str((cur_fr + 1)/settings.lottie_format["fr"]) + "s"
				if val_after:
					new[0].attrib["value"] = "true"
				else:
					new[0].attrib["value"] = "false"
				anim.insert(i+1, new)
				i += 1

		i += 1

	# Make the animation constant
	for waypoint in anim:
		waypoint.attrib["before"] = waypoint.attrib["after"] = "constant"




def get_animated_time_list(child, time_list):
	"""
	Appends all the frames corresponding to the waypoints in the
	animated(child[0]) list, in time_list

	Args:
		child     (lxml.etree._Element) : Parent Element of animation
		time_list (set)                 : Will store all the frames at which waypoints are present

	Returns:
		(None)
	"""
	animated = child[0]
	is_animate = is_animated(animated)
	if is_animate in {settings.NOT_ANIMATED, settings.SINGLE_WAYPOINT}:
		return
	for waypoint in animated:
		frame = get_frame(waypoint)
		time_list.add(frame)


def copy_tcb_average(new_waypoint, waypoint, next_waypoint):
	"""
	Copies the average of TCB values of waypoint and next_waypoint to
	'new_waypoint'
	This is just a way around to determine TCB values at in-between intervals,
	where new waypoints are introduced. Technically we will first calculate the
	tangents at those new waypoints and then calculate any random TCB values
	from those tangents: IMPROVEMENT

	Args:
		new_waypoint  (lxml.etree._Element) : Waypoint at which the values will be averaged
		waypoint      (lxml.etree._Element) : Waypoint contributing in average
		next_waypoint (lxml.etree._Element) : Waypoint contributing in average

	Returns:
		(None)
	"""
	tens, bias, cont = 0, 0, 0      # default values
	tens1, bias1, cont1 = 0, 0, 0   # default values
	if "tension" in waypoint.keys():
		tens = float(waypoint.attrib["tension"])
	if "continuity" in waypoint.keys():
		cont = float(waypoint.attrib["continuity"])
	if "bias" in waypoint.keys():
		bias = float(waypoint.attrib["bias"])
	if "tension" in next_waypoint.keys():
		tens1 = float(next_waypoint.attrib["tension"])
	if "continuity" in next_waypoint.keys():
		cont1 = float(next_waypoint.attrib["continuity"])
	if "bias" in next_waypoint.keys():
		bias1 = float(next_waypoint.attrib["bias"])
	f_tens, f_bias, f_cont = (tens + tens1) / 2.0, (bias + bias1) / 2.0, (cont + cont1) / 2.0
	new_waypoint.attrib["tension"] = str(f_tens)
	new_waypoint.attrib["continuity"] = str(f_cont)
	new_waypoint.attrib["bias"] = str(f_bias)


def copy_tcb(new_waypoint, waypoint):
	"""
	Copies the tension, bias and continuity values from 'waypoint' to
	'new_waypoint'

	Args:
		new_waypoint (lxml.etree._Element) : Waypoint at which the values will be copied
		waypoint     (lxml.etree._Element) : Waypoint to be copied from

	Returns:
		(None)
	"""
	tens, bias, cont = 0, 0, 0      # default values
	if "tension" in waypoint.keys():
		tens = float(waypoint.attrib["tension"])
	if "continuity" in waypoint.keys():
		cont = float(waypoint.attrib["continuity"])
	if "bias" in waypoint.keys():
		bias = float(waypoint.attrib["bias"])
	new_waypoint.attrib["tension"] = str(tens)
	new_waypoint.attrib["continuity"] = str(cont)
	new_waypoint.attrib["bias"] = str(bias)



"""
Will store all the functions corresponding to solids in lottie
"""

import sys
import settings
from common.misc import get_color_hex
from common.Count import Count
from helpers.blendMode import get_blend
from helpers.transform import gen_helpers_transform
from effects.fill import gen_effects_fill
from synfig.group import get_additional_width, get_additional_height
sys.path.append("..")


def gen_layer_solid(lottie, layer, idx):
    """
    Generates the dictionary corresponding to layers/solid.json

    Args:
        lottie (dict)       : Lottie generated solid layer stored here
        layer  (common.Layer.Layer) : Synfig format solid layer
        idx    (int)        : Stores the index(number of) of solid layer

    Returns:
        (None)
    """
    index = Count()
    lottie["ddd"] = settings.DEFAULT_3D
    lottie["ind"] = idx
    lottie["ty"] = settings.LAYER_SOLID_TYPE
    lottie["nm"] = layer.get_description()
    lottie["sr"] = settings.LAYER_DEFAULT_STRETCH
    lottie["ks"] = {}   # Transform properties to be filled
    lottie["ef"] = []   # Stores the effects

    pos = [settings.lottie_format["w"]/2 + get_additional_width()/2,
           settings.lottie_format["h"]/2 + get_additional_height()/2]

    anchor = pos
    gen_helpers_transform(lottie["ks"], pos, anchor)

    lottie["ef"].append({})
    gen_effects_fill(lottie["ef"][-1], layer, index.inc())

    lottie["ao"] = settings.LAYER_DEFAULT_AUTO_ORIENT
    lottie["sw"] = settings.lottie_format["w"] + get_additional_width() # Solid Width
    lottie["sh"] = settings.lottie_format["h"] + get_additional_height() # Solid Height

    lottie["sc"] = get_color_hex(layer.get_param("color").get()[0])

    lottie["ip"] = settings.lottie_format["ip"]
    lottie["op"] = settings.lottie_format["op"]
    lottie["st"] = 0            # Don't know yet
    get_blend(lottie, layer)



"""
Gradient.py
Will store the gradient class
"""

import sys
import copy
import math
import common
import settings
from common.Color import Color
sys.path.append("..")


class Gradient:
    """
    To store the gradients in Synfig and operations on them
    Currently, it is assumed that only 2 colors will be provided in the gradient
    """

    def __init__(self, gradient):
        """
        Args:
            gradient (lxml.etree._Element) : Gradient as extracted from the xml

        Returns:
            (None)
        """
        self.synfig_gradient = gradient
        self.colors = []
        self.extract_colors()

    def extract_colors(self):
        """
        Helps in extracting all the colors present in the gradient
        """
        initial_list = []
        for col in self.synfig_gradient:
            red = float(col[0].text) ** (1/settings.GAMMA[0])
            green = float(col[1].text) ** (1/settings.GAMMA[1])
            blue = float(col[2].text) ** (1/settings.GAMMA[2])
            alpha = float(col[3].text)
            initial_list.append([float(col.attrib["pos"]), Color(red, green, blue, alpha)]) 
        self.colors = initial_list

    def reverse_gamma(self, color):
        """
        Given a color, it reverses the effect of gamma as done in extract_colors()

        Args:
            color (common.Color.Color) : color element

        Returns:
            (common.Color.Color) : color element with gamma effect reversed
        """
        ret = copy.deepcopy(color)
        ret.red = ret.red ** settings.GAMMA[0]
        ret.green = ret.green ** settings.GAMMA[1]
        ret.blue = ret.blue ** settings.GAMMA[2]
        return ret

    def get_lottie_final_list(self):
        """
        Appends the avg of colors as required by lottie format. This function is not currently used
        anywhere. But might be helpful in future

        Args:
            (None)

        Returns:
            (list) : list with average of colors appended
        """
        initial_list = copy.deepcopy(self.colors)
        avg_list = []
        # Appending the avg colors, as required by lottie
        for i in range(len(initial_list)-1):
            col1, col2 = initial_list[i], initial_list[i+1]
            avg_pos = (col1[0] + col2[0]) / 2
            avg_col = (col1[1] + col2[1]) / 2
            avg_list.append([avg_pos, avg_col])

        ret = [None]*(len(initial_list) + len(avg_list))
        ret[::2] = initial_list
        ret[1::2] = avg_list
        return ret

    def get_colors(self):
        """
        Returns the list of colors in a gradient

        Args:
            (None)

        Returns:
            (list) : list of colors of whom gradient is comprised
        """
        return self.colors


    def get_val(self):
        """
        Get the gradient in the format required by lottie

        Args:
            (None)

        Returns:
            (list) : Stores the gradient in the list format
        """
        ret = []
        for col in self.colors:
            pos = col[0]
            r, g, b = col[1].red, col[1].green, col[1].blue
            ret.extend([pos, r, g, b])

        # Alpha values of the colors to be fit here
        for col in self.colors:
            pos = col[0]
            a = col[1].alpha
            ret.extend([pos, a])

        return ret

    def get_color_at_x(self, x):
        """
        Returns the color at any position x(between 0 and 1)
        https://github.com/synfig/synfig/blob/ae11655a9bba068543be7a5df9090958579de78e/synfig-core/src/synfig/gradient.cpp#L178

        Args:
            x (float) : position at which value of color is needed

        Returns:
            (common.Color.Color) : color at position x
        """
        if len(self.colors) == 0:
            return Color()
        if len(self.colors) == 1 or math.isnan(x):
            return self.reverse_gamma(self.colors[0][1])

        # point is outside of gradient
        if x <= self.colors[0][0]:
            return self.reverse_gamma(self.colors[0][1])
        if x >= self.colors[-1][0]:
            return self.reverse_gamma(self.colors[-1][1])

        # upper bound
        i = self.upper_bound(self.colors, x, 0, len(self.colors))
        # considering i stores the upper bound
        j = i
        i = i - 1
        d = self.colors[j][0] - self.colors[i][0]
        if d <= common.misc.real_high_precision():
            return self.reverse_gamma(self.colors[i][1])

        amount = (x - self.colors[i][0]) / d
        col = self.blend(self.reverse_gamma(self.colors[i][1]), self.reverse_gamma(self.colors[j][1]), amount)
        return col

    def blend(self, col1, col2, amount):
        """
        https://github.com/synfig/synfig/blob/ae11655a9bba068543be7a5df9090958579de78e/synfig-core/src/synfig/gradient.cpp#L200

        Args:
            col1 (common.Color.Color) : Color 1
            col2 (common.Color.Color) : Color 2
            amount (float)            : the percentage of what color 1 and color 2 is needed

        Returns:
            (common.Color.Color) : The blended color by some amount 'amount'
        """
        return col1*(1-amount)+col2*amount

    def upper_bound(self, a, x, lo=0, hi=None):
        """
        Function to find the upper bound in an array/list

        Args:
            a (list)  : The list to be searched in
            x (float) : The number whose upper bound is to be found
            lo (int)  : Starting position of the list
            hi (int)  : Ending position of the list

        Returns:
            (int) : The index of upper bound of a number x
        """
        if lo < 0:
            raise ValueError('lo must be non-negative')
        if hi is None:
            hi = len(a)
        while lo < hi:
            mid = (lo+hi)//2
            # Use _lt_ to match the logic in list.sort() and in heapq
            if self.comp(x, a[mid][0]): hi = mid
            else: lo = mid+1
        return lo

    def comp(self, a, b):
        """
        Compares 2 values

        Args:
            a (float) : Number 1 to be compared
            b (float) : Number 2 to be compared

        Returns:
            (bool) : True if b > a under the real_high_precision
                   : False otherwise
        """
        if (b - a) > common.misc.real_high_precision():     # returns True if b > a
            return True
        return False



"""
Implements all the functions required for generating value for properties
related to layers in Synfig
"""

import sys
import settings
sys.path.append("../")


def gen_properties_value(lottie, val, index, animated, expression):
    """
    Generates the dictionary corresponding to properties/value.json in lottie
    documentation

    Args:
        lottie   (dict)                : Value of property is stored in lottie format
        val      (list | float)        : actual value of parameter
        index    (int)                 : Index/Count of value
        animated (int)                 : Holds if this property is animated or not
        expression (str)               : purpose of this is not clearly stated in lottie

    Returns:
        (None)
    """
    lottie["k"] = val
    lottie["ix"] = index
    lottie["a"] = animated
    if expression != settings.NO_INFO:
        lottie["x"] = expression



# pylint: disable=line-too-long
"""
This module will store all the functions required for transformations
corresponding to lottie
"""

import sys
import settings
from common.Count import Count
from properties.value import gen_properties_value
sys.path.append("../")


def gen_helpers_transform(lottie, pos=settings.DEFAULT_POSITION, anchor=settings.DEFAULT_ANCHOR, scale=settings.DEFAULT_SCALE, rotation=settings.DEFAULT_ROTATION, opacity=settings.DEFAULT_OPACITY, skew=settings.DEFAULT_SKEW):
    """
    Generates the dictionary corresponding to helpers/transform.json

    Args:
        lottie (dict)                : Lottie format layer
        pos    (:obj: `list | lxml.etree._Element`, optional) : position of layer
        anchor (:obj: `list | lxml.etree._Element`, optional) : anchor point of layer
        scale (:obj: `list | lxml.etree._Element`, optional) : scale of layer
        rotation (:obj: `float | lxml.etree._Element`, optional) : rotation of layer
        opacity (:obj: `float | lxml.etree._Element`, optional) : Opacity of layer

    Returns:
        (None)
    """
    index = Count()
    lottie["o"] = {}    # opacity/Amount
    lottie["r"] = {}    # Rotation of the layer
    lottie["p"] = {}    # Position of the layer
    lottie["a"] = {}    # Anchor point of the layer
    lottie["s"] = {}    # Scale of the layer
    lottie["sk"] = {}   # skew of the layer
    lottie["sa"] = {}   # skew axis of the layer

    # setting the position
    if isinstance(pos, list):
        gen_properties_value(lottie["p"],
                             pos,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)
    else:
        pos.fill_path(lottie, "p")

    # setting the opacity
    if isinstance(opacity, (float, int)):
        gen_properties_value(lottie["o"],
                             opacity,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)
    else:
        opacity.fill_path(lottie, "o")

    # setting the rotation
    if isinstance(rotation, (float, int)):
        gen_properties_value(lottie["r"],
                             rotation,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)
    else:
        rotation.fill_path(lottie, "r")

    # setting the anchor point
    if isinstance(anchor, list):
        gen_properties_value(lottie["a"],
                             anchor,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)
    else:
        anchor.fill_path(lottie, "a")

    # setting the scale
    if isinstance(scale, list):
        gen_properties_value(lottie["s"],
                             scale,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)
    # This means scale parameter is animated
    else:
        scale.fill_path(lottie, "s")

    # setting the skew angle
    if isinstance(skew, (float, int)):
        gen_properties_value(lottie["sk"],
                             skew,
                             index.inc(),
                             settings.DEFAULT_ANIMATED,
                             settings.NO_INFO)
    else:
        skew.fill_path(lottie, "sk")

    # setting the skew axis
    gen_properties_value(lottie["sa"],
                         0,
                         index.inc(),
                         settings.DEFAULT_ANIMATED,
                         settings.NO_INFO)



"""
This module will store all the functions required for all mask property of lottie
"""

import sys
import settings
from common.Count import Count
from properties.value import gen_properties_value
sys.path.append("../")


def gen_effects_allmask(lottie, idx):
    """
    Generates the dictionary corresponding to effects/allmask.json

    Args:
        lottie (dict)                : Lottie format effects stored in this
        idx    (int)                 : Index/Count of effect

    Returns:
        (None)
    """
    index = Count()
    lottie["ty"] = settings.EFFECTS_ALL_MASK        # Effect type
    lottie["nm"] = "All Masks"                      # Name
    lottie["ix"] = idx                              # Index
    lottie["v"] = {}                                # value
    gen_properties_value(lottie["v"], 0, index.inc(), 0, settings.NO_INFO)



"""
This module will store all the functions required for color property of lottie
"""

import sys
import settings
sys.path.append("../")


def gen_effects_color(lottie, layer, idx):
    """
    Generates the dictionary corresponding to effects/color.json

    Args:
        lottie (dict)               : Lottie format effects stored in this
        layer  (common.Layer.Layer)  : Synfig format layer
        idx    (int)                : Index/Count of effect

    Returns:
        (None)

    """
    lottie["ty"] = settings.EFFECTS_COLOR   # Effect type
    lottie["nm"] = "Color"                  # Name
    lottie["ix"] = idx                      # Index
    lottie["v"] = {}                        # Value of color

    color = layer.get_param("color")
    color.animate("color")
    color.fill_path(lottie, "v")



# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of outline layer
in Lottie format
"""

import sys
import math
import copy
import settings
from common.Bline import Bline
from common.Vector import Vector
from common.Hermite import Hermite
from synfig.animation import to_Synfig_axis
from properties.shapePropKeyframe.helper import add_reverse, add, move_to, insert_dict_at, animate_tangents
sys.path.append("../../")


def gen_bline_outline(lottie, bline_point):
    """
    Generates the bline corresponding to outline layer by adding some vertices
    to bline and converting it to region layer

    Some parts are common with gen_bline_shapePropKeyframe(), which will be
    placed in a common function latter

    Args:
        lottie (dict) : Lottie format outline layer will be stored in this
        bline_point (common.Param.Param) : Synfig format bline points

    Returns:
        (None)
    """
    ################### SECTION 1 #########################
    # Inserting waypoints if not animated and finding the first and last frame
    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1

    bline = Bline(bline_point[0], bline_point)

    for entry in bline.get_entry_list():
        pos = entry["point"]
        origin = entry["origin"]
        width = entry["width"]
        t1 = entry["t1"]
        t2 = entry["t2"]
        split_r = entry["split_radius"]
        split_a = entry["split_angle"]

        pos.update_frame_window(window)
        # Empty the pos and fill in the new animated pos
        pos.animate("vector")

        origin.update_frame_window(window)
        origin.animate("real")

        width.update_frame_window(window)
        width.animate("real")

        split_r.update_frame_window(window)
        split_r.animate_without_path("bool")

        split_a.update_frame_window(window)
        split_a.animate_without_path("bool")

        animate_tangents(t1, window)
        animate_tangents(t2, window)

        # Open up the window if bline points are animated
        entry["ActivepointList"].update_frame_window(window)

    layer = bline.get_layer().get_layer()
    outer_width = layer.get_param("width")
    sharp_cusps = layer.get_param("sharp_cusps")
    expand = layer.get_param("expand")
    r_tip0 = layer.get_param("round_tip[0]")
    r_tip1 = layer.get_param("round_tip[1]")
    homo_width = layer.get_param("homogeneous_width")
    origin = layer.get_param("origin")

    # Animating the origin
    origin.update_frame_window(window)
    origin.animate("vector")

    # Animating the outer width
    outer_width.update_frame_window(window)
    outer_width.animate("real")

    # Animating the sharp_cusps
    sharp_cusps.update_frame_window(window)
    sharp_cusps.animate_without_path("bool")

    # Animating the expand param
    expand.update_frame_window(window)
    expand.animate("real")

    # Animating the round tip 0
    r_tip0.update_frame_window(window)
    r_tip0.animate_without_path("bool")

    # Animating the round tip 1
    r_tip1.update_frame_window(window)
    r_tip1.animate_without_path("bool")

    # Animating the homogeneous width
    homo_width.update_frame_window(window)
    homo_width.animate_without_path("bool")

    # Minimizing the window size
    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    ################# END OF SECTION 1 ###################

    ################ SECTION 2 ###########################
    # Generating values for all the frames in the window


    lottie_st_list, lottie_en_list = [], []
    fr = window["first"]
    while fr <= window["last"]:
        st_val, en_val = insert_dict_at(lottie, -1, fr, False)  # This loop needs to be considered somewhere down
        lottie_st_list.append(st_val)
        lottie_en_list.append(en_val)
        synfig_outline(bline, st_val, origin, outer_width, sharp_cusps, expand, r_tip0, r_tip1, homo_width, fr)
        synfig_outline(bline, en_val, origin, outer_width, sharp_cusps, expand, r_tip0, r_tip1, homo_width, fr + 1)
        fr += 1
    equalize_length(lottie_st_list, lottie_en_list)
    # Setting the final time
    lottie.append({})
    lottie[-1]["t"] = fr


def equalize_length(lottie_st, lottie_en):
    """
    This fxn collects all the lists needed to render the outline, and makes them
    of equal length by adding dummy points: This is a requirement of Lottie
    """
    # Find maximum among the lists
    mx = 0
    for st in lottie_st:
        mx = max(mx, len(st["i"]))
    for en in lottie_en:
        mx = max(mx, len(en["i"]))

    for st in lottie_st:
        diff = mx - len(st["i"])
        last_value_i, lv_o, lv_v = copy.deepcopy(st["i"][-1]), copy.deepcopy(st["o"][-1]), copy.deepcopy(st["v"][-1])
        st["i"].extend([last_value_i for i in range(diff)])
        st["o"].extend([lv_o for i in range(diff)])
        st["v"].extend([lv_v for i in range(diff)])

    for st in lottie_en:
        diff = mx - len(st["i"])
        last_value_i, lv_o, lv_v = copy.deepcopy(st["i"][-1]), copy.deepcopy(st["o"][-1]), copy.deepcopy(st["v"][-1])
        st["i"].extend([last_value_i for i in range(diff)])
        st["o"].extend([lv_o for i in range(diff)])
        st["v"].extend([lv_v for i in range(diff)])


def get_outline_grow(fr):
    """
    Gives the value of outline grow parameter at a particular frame
    """
    ret = 0
    for og in settings.OUTLINE_GROW:
        if isinstance(og, (float, int)):
            ret += og
        else:   # should be dict
            val = to_Synfig_axis(og.get_value(fr), "real")
            ret += val
    ret = math.e ** ret
    return ret


def synfig_outline(bline, st_val, origin_p, outer_width_p, sharp_cusps_p, expand_p, r_tip0_p, r_tip1_p, homo_width_p, fr):
    """
    Calculates the points for the outline layer as in Synfig:
    https://github.com/synfig/synfig/blob/678cc3a7b1208fcca18c8b54a29a20576c499927/synfig-core/src/modules/mod_geometry/outline.cpp

    Args:
        bline_point (common.Bline.Bline) : Synfig format bline points of outline layer
        st_val (dict) : Lottie format outline stored in this
        origin_p (common.Param.Param) : Lottie format origin of outline layer
        outer_width_p (common.Param.Param) : Lottie format outer width
        sharp_cusps_p (common.Param.Param) : sharp cusps in Synfig format
        expand_p (common.Param.Param) : Lottie format expand parameter
        r_tip0_p (common.Param.Param) : Round tip[0] in Synfig format
        r_tip1_p (common.Param.Param) : Round tip[1] in Synfig format
        homo_width_p (common.Param.Param) : Homogeneous width in Synfig format
        fr (int) : Frame number

    Returns:
        (None)
    """

    EPSILON = 0.000000001
    SAMPLES = 50
    CUSP_TANGENT_ADJUST = 0.025
    CUSP_THRESHOLD = 0.40
    SPIKE_AMOUNT = 4
    ROUND_END_FACTOR = 4

    bline_list = bline.get_list_at_frame(fr)
    outer_width = to_Synfig_axis(outer_width_p.get_value(fr), "real")
    expand = to_Synfig_axis(expand_p.get_value(fr), "real")
    sharp_cusps = sharp_cusps_p.get_value(fr)
    r_tip0 = r_tip0_p.get_value(fr)
    r_tip1 = r_tip1_p.get_value(fr)
    homo_width = homo_width_p.get_value(fr)

    gv = get_outline_grow(fr)

    # Setup chunk list
    side_a, side_b = [], []

    # Check if looped
    loop = bline.get_loop()

    # Iterators
    end_it = len(bline_list)
    next_it = 0
    if loop:
        iter_it = end_it - 1
    else:
        iter_it = next_it
        next_it += 1

    first_point = bline_list[iter_it]
    first_tangent = bline_list[0].get_tangent2()
    last_tangent = first_point.get_tangent1()

    # If we are looped and drawing sharp cusps, we 'll need a value
    # for the incoming tangent. This code fails if we have a "degraded" spline
    # with just one vertex, so we avoid such case.
    if loop and sharp_cusps and last_tangent.is_equal_to(Vector(0, 0)) and len(bline_list) > 1:
        prev_it = iter_it
        prev_it -= 1
        prev_it %= len(bline_list)
        prev_point = bline_list[prev_it]
        curve = Hermite(prev_point.get_vertex(),
                        first_point.get_vertex(),
                        prev_point.get_tangent2(),
                        first_point.get_tangent1())
        last_tangent = curve.derivative(1.0 - CUSP_TANGENT_ADJUST)

    first = not loop
    while next_it != end_it:
        bp1 = bline_list[iter_it]
        bp2 = bline_list[next_it]

        # Setup tangents
        prev_t = bp1.get_tangent1()
        iter_t = bp1.get_tangent2()
        next_t = bp2.get_tangent1()

        split_flag = bp1.get_split_tangent_angle() or bp1.get_split_tangent_radius()

        # If iter_it.t2 == 0 and next.t1 == 0, this is a straight line
        if iter_t.is_equal_to(Vector(0, 0)) and next_t.is_equal_to(Vector(0, 0)):
            iter_t = next_t = bp2.get_vertex() - bp1.get_vertex()

            # If the 2 points are on top of each other, ignore this segment
            # leave 'first' true if was before
            if iter_t.is_equal_to(Vector(0, 0)):
                iter_it = next_it
                iter_it %= len(bline_list)
                next_it += 1
                continue

        # Setup the curve
        curve = Hermite(bp1.get_vertex(),
                        bp2.get_vertex(),
                        iter_t,
                        next_t)

        # Setup width's
        iter_w = gv*(bp1.get_width() * outer_width * 0.5 + expand)
        next_w = gv*(bp2.get_width() * outer_width * 0.5 + expand)

        if first:
            first_tangent = curve.derivative(CUSP_TANGENT_ADJUST)

        # Make cusps as necassary
        if not first and \
           sharp_cusps and \
           split_flag and \
           ((not prev_t.is_equal_to(iter_t)) or (iter_t.is_equal_to(Vector(0, 0)))) and \
           (not last_tangent.is_equal_to(Vector(0, 0))):

            curr_tangent = curve.derivative(CUSP_TANGENT_ADJUST)
            t1 = last_tangent.perp().norm()
            t2 = curr_tangent.perp().norm()

            cross = t1*t2.perp()
            perp = (t1 - t2).mag()
            if cross > CUSP_THRESHOLD:
                p1 = bp1.get_vertex() + t1*iter_w
                p2 = bp1.get_vertex() + t2*iter_w
                side_a.append([line_intersection(p1, last_tangent, p2, curr_tangent), Vector(0, 0), Vector(0, 0)])
            elif cross < -CUSP_THRESHOLD:
                p1 = bp1.get_vertex() - t1*iter_w
                p2 = bp1.get_vertex() - t2*iter_w
                side_b.append([line_intersection(p1, last_tangent, p2, curr_tangent), Vector(0, 0), Vector(0, 0)])
            elif cross > 0.0 and perp > 1.0:
                amount = max(0.0, cross/CUSP_THRESHOLD) * (SPIKE_AMOUNT - 1.0) + 1.0
                side_a.append([bp1.get_vertex() + (t1 + t2).norm()*iter_w*amount, Vector(0, 0), Vector(0, 0)])
            elif cross < 0 and perp > 1:
                amount = max(0.0, -cross/CUSP_THRESHOLD) * (SPIKE_AMOUNT - 1.0) + 1.0
                side_b.append([bp1.get_vertex() - (t1 + t2).norm()*iter_w*amount, Vector(0, 0), Vector(0, 0)])

        # Precalculate positions and coefficients
        length = 0.0
        points = []
        dists = []
        n = 0.0
        itr = 0
        while n < 1.000001:
            points.append(curve.value(n))
            if n:
                length += (points[itr] - points[itr-1]).mag()
            dists.append(length)

            n += 1.0/SAMPLES
            itr += 1
        length += (curve.value(1) - points[itr-1]).mag()

        div_length = 1
        if length > EPSILON:
            div_length = 1.0 / length

        # Might not need /3 for the tangents generated finally - VERY IMPORTANT
        # Make the outline
        pt = curve.derivative(CUSP_TANGENT_ADJUST) / 3
        n = 0.0
        itr = 0
        while n < 1.000001:
            t = curve.derivative(min(max(n, CUSP_TANGENT_ADJUST), 1.0 - CUSP_TANGENT_ADJUST)) / 3
            d = t.perp().norm()
            k = dists[itr] * div_length
            if not homo_width:
                k = n
            w = (next_w - iter_w)*k + iter_w
            if False and n:
                # create curve
                a = points[itr-1] + d*w
                b = points[itr] + d*w
                tk = (b - a).mag() * div_length
                side_a.append([b, pt*tk, -t*tk])

                a = points[itr-1] - d*w
                b = points[itr] - d*w
                tk = (b - a).mag() * div_length
                side_b.append([b, pt*tk, -t*tk])
            else:
                side_a.append([points[itr] + d*w, Vector(0, 0), Vector(0, 0)])
                side_b.append([points[itr] - d*w, Vector(0, 0), Vector(0, 0)])
            pt = t
            itr += 1
            n += 1.0/SAMPLES

        last_tangent = curve.derivative(1.0 - CUSP_TANGENT_ADJUST)
        side_a.append([curve.value(1.0) + last_tangent.perp().norm()*next_w, Vector(0, 0), Vector(0, 0)])
        side_b.append([curve.value(1.0) - last_tangent.perp().norm()*next_w, Vector(0, 0), Vector(0, 0)])
        first = False

        iter_it = next_it
        iter_it %= len(bline_list)
        next_it += 1

    if len(side_a) < 2 or len(side_b) < 2:
        return

    origin_cur = origin_p.get_value(fr)
    move_to(side_a[0][0], st_val, origin_cur)

    if loop:
        add(side_a, st_val, origin_cur)
        add_reverse(side_b, st_val, origin_cur)
    else:
        # Insert code for adding end tip
        if r_tip1:
            bp = bline_list[-1]
            vertex = bp.get_vertex()
            tangent = last_tangent.norm()
            w = gv*(bp.get_width() * outer_width * 0.5 + expand)

            a = vertex + tangent.perp()*w
            b = vertex - tangent.perp()*w
            p1 = a + tangent*w*(ROUND_END_FACTOR/3.0)
            p2 = b + tangent*w*(ROUND_END_FACTOR/3.0)
            tan = tangent*w*(ROUND_END_FACTOR/3.0)

            # replace the last point
            side_a[-1] = [a, Vector(0, 0), tan]
            add(side_a, st_val, origin_cur)
            add([[b, -tan, Vector(0, 0)]], st_val, origin_cur)
        else:
            add(side_a, st_val, origin_cur)

        # Insert code for adding beginning tip
        if r_tip0:
            bp = bline_list[0]
            vertex = bp.get_vertex()
            tangent = first_tangent.norm()
            w = gv*(bp.get_width() * outer_width * 0.5 + expand)

            a = vertex - tangent.perp()*w
            b = vertex + tangent.perp()*w
            p1 = a - tangent*w*(ROUND_END_FACTOR/3.0)
            p2 = b - tangent*w*(ROUND_END_FACTOR/3.0)
            tan = -tangent*w*(ROUND_END_FACTOR/3.0)

            # replace the first point
            side_b[0] = [a, Vector(0, 0), tan]
            add_reverse(side_b, st_val, origin_cur)
            add([[b, -tan, Vector(0, 0)]], st_val, origin_cur)
        else:
            add_reverse(side_b, st_val, origin_cur)


def line_intersection(p1, t1, p2, t2):
    """
    This function was adapted from what was
    described on http://www.whisqu.se/per/docs/math28.htm

    Args:
        p1 (common.Vector.Vector) : First point
        t1 (common.Vector.Vector) : First tangent
        p2 (common.Vector.Vector) : Second point
        t2 (common.Vector.Vector) : Second tangent

    Returns:
        (common.Vector.Vector) : intersection of the both the lines
    """
    x0 = p1[0]
    y0 = p1[1]

    x1 = p1[0] + t1[0]
    y1 = p1[1] + t1[1]

    x2 = p2[0]
    y2 = p2[1]

    x3 = p2[0] + t2[0]
    y3 = p2[1] + t2[1]

    near_infinity = 1e+10

    # compute slopes, not the kluge for infinity, however, this will
    # be close enough

    if (x1 - x0) != 0:
        m1 = (y1 - y0) / (x1 - x0)
    else:
        m1 = near_infinity

    if (x3 - x2) != 0:
        m2 = (y3 - y2) / (x3 - x2)
    else:
        m2 = near_infinity

    a1 = m1
    a2 = m2
    b1 = -1.0
    b2 = -1.0
    c1 = y0 - m1*x0
    c2 = y2 - m2*x2

    # compute the inverse of the determinate
    if a1*b2 - a2*b1 == 0:
        det_inv = float('nan')
    else:
        det_inv = 1.0 / (a1*b2 - a2*b1)

    # Use kramer's rule to compute intersection
    return Vector((b1*c2 - b2*c1)*det_inv, (a2*c1 - a1*c2)*det_inv)



# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of star layer
in Lottie format
"""

import sys
import math
from common.Vector import Vector
from synfig.animation import to_Synfig_axis
from properties.shapePropKeyframe.helper import add, insert_dict_at
sys.path.append("../../")


def gen_list_star(lottie, layer):
    """
    Generates a shape layer corresponding to star layer by manipulating the
    parameters of the star

    Args:
        lottie (dict) : Lottie format rectangle layer will be stored in this
        layer  (common.Layer.Layer) : Synfig format rectangle layer

    Returns:
        (None)
    """
    ################### SECTION 1 #########################
    # Inserting waypoints if not animated and finding the first and last frame
    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1

    origin = layer.get_param("origin")
    radius1 = layer.get_param("radius1")
    radius2 = layer.get_param("radius2")
    angle = layer.get_param("angle")
    points = layer.get_param("points")
    regular_polygon = layer.get_param("regular_polygon")

    # Animating origin
    origin.update_frame_window(window)
    # Keeping the transform true here
    #origin.animate("vector", True)
    origin.animate("vector")

    # Animating radius1
    radius1.update_frame_window(window)
    radius1.animate("real")

    # Animating radius2
    radius2.update_frame_window(window)
    radius2.animate("real")

    # Animating angle
    angle.update_frame_window(window)
    angle.animate("star_angle_new")

    # Animating points
    points.update_frame_window(window)
    points.animate("real")

    mx_points = get_max_points(points)

    # Animating regular_polygon
    regular_polygon.update_frame_window(window)
    regular_polygon.animate_without_path("bool")

    # Minimizing the window size
    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    ################# END OF SECTION 1 ###################

    ################ SECTION 2 ###########################
    # Generating values for all the frames in the window

    fr = window["first"]
    while fr <= window["last"]:
        st_val, en_val = insert_dict_at(lottie, -1, fr, False)
        synfig_star(st_val, mx_points, origin, radius1, radius2, angle, points, regular_polygon, fr)
        synfig_star(en_val, mx_points, origin, radius1, radius2, angle, points, regular_polygon, fr + 1)

        fr += 1
    # Setting the final time
    lottie.append({})
    lottie[-1]["t"] = fr


def get_max_points(points):
    """
    As a shape layer needs to have same number of vertices at each frame, we
    will draw a shape with max number of points. When number of points at a
    frame is less than mx, redundant vertices on the shape will be added

    NOTE/IMPORTANT: There might come a case when somewhere between two
    waypoints, the number of points is greater than both those 2 waypoints(very
    rare). This case is not handled here

    Args:
        points (lxml.etree._Element) : Animation of points param in Synfig format

    Returns:
        mx (int) : Maximum number of points ever in the animation
    """
    mx = -1
    for waypoint in points[0]:
        val = float(waypoint[0].attrib["value"])
        mx = max(mx, val)
    return mx


def synfig_star(st_val, mx_points, origin_p, radius1_p, radius2_p, angle_p, points_p, regular_polygon_p, fr):
    """
    Calculates the points for the rectangle layer as in Synfig:
    https://github.com/synfig/synfig/blob/678cc3a7b1208fcca18c8b54a29a20576c499927/synfig-core/src/modules/mod_geometry/star.cpp

    Args:
        st_val (dict) : Lottie format star will be stored here
        mx_points (int) : Maximum points ever in star animation
        radius1_p (common.Param.Param) : Lottie format radius1 animation
        radius2_p (common.Param.Param) : Lottie format radius2 animation
        angle_p   (common.Param.Param) : Lottie format angle animation
        points_p  (common.Param.Param) : Lottie format points animation
        regular_polygon_p (common.Param.Param) : Synfig format regularPolygon animation
        fr (int) : Frame number

    Returns:
        (None)
    """

    angle = angle_p.get_value(fr)
    points = int(to_Synfig_axis(points_p.get_value(fr), "real"))
    radius1 = to_Synfig_axis(radius1_p.get_value(fr), "real")
    radius2 = to_Synfig_axis(radius2_p.get_value(fr), "real")
    regular_polygon = regular_polygon_p.get_value(fr)
    origin_cur = origin_p.get_value(fr)

    angle = math.radians(angle)
    dist_between_points = (math.pi * 2) / float(points)
    vector_list = []

    ####
    tot_points = 2*mx_points

    i = 0
    while i < points:
        dist1 = dist_between_points*i + angle
        dist2 = dist_between_points*i + dist_between_points/2 + angle
        vector_list.append(Vector(math.cos(dist1)*radius1, math.sin(dist1)*radius1))
        if not regular_polygon:
            vector_list.append(Vector(math.cos(dist2)*radius2, math.sin(dist2)*radius2))
        else:
            # This condition is needed because in lottie a shape must have equal
            # number of vertices at each frame
            vector_list.append(Vector(math.cos(dist1)*radius1, math.sin(dist1)*radius1))

        tot_points -= 2
        i += 1

    if len(vector_list) < 3:
        # Should throw error
        return

    while tot_points > 0:
        vector_list.append(vector_list[-1])
        tot_points -= 1

    # Setup chunk list
    chunk_list = []
    chunk_list.append([vector_list[0], Vector(), Vector()])
    i = 1
    while i < len(vector_list):
        if math.isnan(vector_list[i][0]) or math.isnan(vector_list[i][1]):
            break
        chunk_list.append([vector_list[i], Vector(), Vector()])
        i += 1

    add(chunk_list, st_val, origin_cur)



"""
Color.py
Will store the Color class

NOTE: need to introduce clamping function
"""

import sys
sys.path.append("..")


class Color:
    """
    To store the colors in Synfig and operations on them
    """

    def __init__(self, red=1, green=1, blue=1, alpha=1):
        """
        Args:
            red (:obj: `float`, optional) : Red value of color
            green (:obj: `float`, optional) : Green value of color
            blue (:obj: `float`, optional) : Blue value of color
            alpha (:obj: `float`, optional) : Alpha value of color

        Returns:
            (None)
        """
        self.red = red
        self.green = green
        self.blue = blue
        self.alpha = alpha

    def __str__(self):
        return "({0}, {1}, {2}, {3})".format(self.red, self.green, self.blue,
                                             self.alpha)

    def __add__(self, other):
        red = self.red + other.red
        green = self.green + other.green
        blue = self.blue + other.blue
        return Color(red, green, blue, self.alpha)

    def __sub__(self, other):
        red = self.red - other.red
        green = self.green - other.green
        blue = self.blue - other.blue
        return Color(red, green, blue, self.alpha)

    def __mul__(self, other):
        if not isinstance(other, self.__class__):
            red = self.red * other
            green = self.green * other
            blue = self.blue * other
            return Color(red, green, blue, self.alpha)
        raise Exception('Multiplication with {} not defined'.format(type(other)))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if not isinstance(other, self.__class__):
            red = self.red / other
            green = self.green / other
            blue = self.blue / other
            return Color(red, green, blue, self.alpha)
        raise Exception('Division with {} not defined'.format(type(other)))

    def get_val(self):
        """
        Get the color in the format required by lottie

        Args:
            (None)

        Returns:
            (list) : Stores color in list format
        """
        return [self.red, self.green, self.blue, self.alpha]



"""
WidthPoint.py
To store a single Width Point, this class will be used

REFER HERE:
https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/widthpoint.cpp
"""

import sys
import math
from functools import total_ordering
sys.path.append("..")


@total_ordering
class WidthPoint:
    """
    Class to store Synfig format Width points
    ****TODO: Synfig is currently not using lower_bound and upper_bound values
    They are set to 0 and 1 respectively
    """
    def __init__(self, position, width, sidebefore, sideafter, dash = False, lower_bound = 0.0, upper_bound = 1.0):
        self.position_ = position
        self.width_ = width
        self.dash_ = dash
        self.side_type_ = []
        self.side_type_.append(sidebefore)
        self.side_type_.append(sideafter)
        self.lower_bound_ = lower_bound
        self.upper_bound_ = upper_bound

    def set_position(self, x):
        self.position_ = x

    def get_norm_position(self, wplistloop):
        return (self.get_bound_position(wplistloop)- self.lower_bound_) / (self.upper_bound_ - self.lower_bound_)

    def normalize(self, loop):
        self.set_position(self.get_norm_position(loop))

    def get_bound_position(self, wplistloop):
        ret = self.position_ - self.lower_bound_
        rnge = self.upper_bound_ - self.lower_bound_
        if not wplistloop:
            if ret < 0: ret = 0
            if ret > rnge: ret = rnge
        else:
            ret = math.fmod(ret, rnge)
            ret = ret if (ret >= 0.0) else (rnge + ret) 
        return ret + self.lower_bound_

    def set_side_type_before(self, sidebefore):
        self.side_type_[0] = sidebefore

    def get_side_type_before(self):
        return self.side_type_[0]

    def set_side_type_after(self, sideafter):
        self.side_type_[1] = sideafter

    def get_side_type_after(self):
        return self.side_type_[1]

    def get_position(self):
        return self.position_

    def set_width(self, width):
        self.width_ = width

    def get_width(self):
        return self.width_

    def get_dash(self):
        return self.dash_

    def __lt__(self, rhs):
        return ((self.position_) < (rhs.position_))

    def __eq__(self, rhs):
        return (self.position_ == rhs.position_ and
                self.width_ == rhs.width_ and
                self.lower_bound_ == rhs.lower_bound_ and
                self.upper_bound_ == rhs.upper_bound_ and
                self.side_type[0] == rhs.side_type[0] and
                self.side_type[1] == rhs.side_type[1])



"""
Will store all the functions required for generation of
multiDimensionalKeyframed file in lottie
"""

import sys
from properties.offsetKeyframe import gen_properties_offset_keyframe
from properties.timeAdjust import time_adjust
from common.misc import get_frame
sys.path.append("..")


def gen_properties_multi_dimensional_keyframed(lottie, animated, idx):
    """
    Generates the dictionary corresponding to
    properties/multiDimensionalKeyframed.json

    Args:
        lottie   (dict)                : Lottie generated keyframes will be stored here
        animated (lxml.etree._Element) : Synfig format animation
        idx      (int)                 : Index/Count of animation

    Returns:
        (None)
    """
    lottie["a"] = 1
    lottie["ix"] = idx
    lottie["k"] = []
    for i in range(len(animated) - 1):
        lottie["k"].append({})
        gen_properties_offset_keyframe(lottie["k"], animated, i)
    last_waypoint_frame = get_frame(animated[-1])
    lottie["k"].append({})
    lottie["k"][-1]["t"] = last_waypoint_frame

    if  "h" in lottie["k"][-2].keys():
        lottie["k"][-1]["h"] = 1
        lottie["k"][-1]["s"] = lottie["k"][-2]["e"]

    # Time adjust of the curves
    time_adjust(lottie, animated)



"""
Bline.py
Will store the Parameters class for Bline
"""

import sys
import copy
import common
import settings
from common.BlinePoint import BlinePoint
from common.Vector import Vector
from common.Hermite import Hermite
from synfig.animation import to_Synfig_axis
from common.Angle import TanAngle, DegAngle, CosAngle, SinAngle
from common.ActivepointList import ActivepointList
from properties.shapePropKeyframe.helper import get_tangent_at_frame
sys.path.append("..")


class Bline:
    """
    Class to store Synfig format bline
    """
    def __init__(self, bline, parent):
        """
        This parent should be parameter
        """
        self.parent = parent
        self.bline = bline
        self.type = bline.attrib["type"]
        self.entry_list = []
        self.extract_entries(self.entry_list)

    def get(self):
        """
        Returns the original param
        """
        return self.bline

    def __getitem__(self, itr):
        """
        Returns the child corresponding to itr
        """
        return self.entry_list[itr]

    def __setitem__(self, itr, val):
        """
        Sets the value of child corresponding to itr
        """
        self.entry_list[itr] = val

    def get_layer_type(self):
        """
        Recursively go till the top until layer is not reached
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent.get_type()
        return self.parent.get_layer_type()

    def get_layer(self):
        """
        Recursively find the layer of this bline
        """
        if isinstance(self.parent, common.Layer.Layer):
            return self.parent
        return self.parent

    def get_type(self):
        """
        Returns the type of bline
        """
        return self.type

    def get_loop(self):
        """
        Returns whether the bline is looped or not
        """
        loop = False
        if "loop" in self.bline.keys():
            v = self.bline.attrib["loop"]
            if v == "true":
                loop = True
        return loop

    def extract_entries(self, entry_list):
        """
        Stores the entries in a list
        """
        for entry in self.bline:
            # Assuming it's child is always composite for now
            entry_list.append({})
            entry_list[-1]["ActivepointList"] = ActivepointList(entry.attrib.get("on"), entry.attrib.get("off"))
            if entry[0].tag == "composite":
                element = entry[0]
            else:
                element = entry
            for param in element:
                tag = param.tag
                if tag == "animated":   # For dynamic list
                    tag = param.attrib["type"]
                entry_list[-1][tag] = common.Param.Param(param, element)
            if entry[0].tag == "composite":
                entry_list[-1]["composite"] = element

    def get_entry_list(self):
        """
        Returns the entry list
        """
        return self.entry_list

    def get_len(self):
        """
        Returns the number of entries
        """
        return len(self.entry_list)

    def set_entry(self, itr, entry, tag="vector"):
        """
        Sets the entry at itr to val
        """
        for param in entry:
            self.entry_list[itr][tag] = common.Param.Param(param, entry)

    def get_list_at_frame_old(self, fr):
        """
        Returns the Bline list at a particular frame
        """
        bline_list = []
        for entry in self.get_entry_list():
            curr = self.get_value_at_frame(entry, fr)
            bline_list.append(curr)
        return bline_list

    def get_value_at_frame(self, entry, fr):
        """
        """
        pos = to_Synfig_axis(entry["point"].get_value(fr), "vector")
        pos = common.Vector.Vector(pos[0], pos[1])
        width = to_Synfig_axis(entry["width"].get_value(fr), "real")
        origin = to_Synfig_axis(entry["origin"].get_value(fr), "real")
        t1 = entry["t1"]
        t2 = entry["t2"]
        split_r = entry["split_radius"]
        split_a = entry["split_angle"]
        t1, t2 = get_tangent_at_frame(t1, t2, fr) 
        # convert to synfig units
        t1 /= settings.PIX_PER_UNIT
        t2 /= settings.PIX_PER_UNIT
        split_r_val = split_r.get_value(fr)
        split_a_val = split_a.get_value(fr)
        return BlinePoint(pos, width, split_r_val, split_a_val, t1, t2, origin)

    def get_list_at_frame(self, fr):
        """
        Returns the Bline list at a particular frame
        Refer: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_bline.cpp
        """
        EPSILON = 0.0000001
        ret_list = []
        first_flag = True
        rising = [False]
        next_scale = 1.0

        first = BlinePoint(Vector(0, 0), 1, True, False, Vector(0, 0), Vector(0, 0))
        first.set_origin(100)

        iterr = 0
        while iterr != self.get_len():
            entry = self.get_entry_list()[iterr]
            amount = entry["ActivepointList"].amount_at_time(fr, rising)
            assert(amount >= 0.0)
            assert(amount <= 1.0)

            # It's fully on
            if amount > 1.0 - EPSILON:
                if first_flag:
                    first_iter = iterr
                    first = prev = self.get_blinepoint(iterr, fr)
                    first_flag = False
                    ret_list.append(copy.deepcopy(first))
                    iterr += 1
                    continue
                curr = self.get_blinepoint(iterr, fr)

                if next_scale != 1.0:
                    ret_list[-1].set_split_tangent_both(True)
                    ret_list[-1].set_tangent2(prev.get_tangent2()*next_scale)
                    ret_list.append(copy.deepcopy(curr))
                    ret_list[-1].set_split_tangent_both(True)
                    ret_list[-1].set_tangent2(curr.get_tangent2())
                    ret_list[-1].set_tangent1(curr.get_tangent1()*next_scale)
                    next_scale = 1.0
                else:
                    ret_list.append(copy.deepcopy(curr))
                prev = curr

            # It's partly on
            elif amount > 0.0:
                blp_here_off = BlinePoint(Vector(0, 0), 1, True, False, Vector(0, 0), Vector(0, 0))
                blp_here_now = BlinePoint(Vector(0, 0), 1, True, False, Vector(0, 0), Vector(0, 0))
                blp_prev_off = BlinePoint(Vector(0, 0), 1, True, False, Vector(0, 0), Vector(0, 0))
                dist_from_begin = 0
                dist_from_end = 0

                if not rising[0]:
                    try:
                        on_time = self.get_entry_list()[iterr]["ActivepointList"].find_prev(fr).get_time()
                    except Exception as e:
                        on_time = settings.SOT
                    try:
                        off_time = self.get_entry_list()[iterr]["ActivepointList"].find_next(fr).get_time()
                    except Exception as e:
                        off_time = settings.EOT
                else:
                    try:
                        off_time = self.get_entry_list()[iterr]["ActivepointList"].find_prev(fr).get_time()
                    except Exception as e:
                        off_time = settings.SOT
                    try:
                        on_time = self.get_entry_list()[iterr]["ActivepointList"].find_next(fr).get_time()
                    except Exception as e:
                        on_time = settings.EOT

                blp_here_on = self.get_blinepoint(iterr, on_time)
                end_iter = iterr

                end_iter += 1
                while end_iter != self.get_len():
                    if self.get_entry_list()[end_iter]["ActivepointList"].amount_at_time(fr) > amount:
                        break
                    end_iter += 1

                if end_iter == self.get_len():
                    if self.get_loop() and (not first_flag):
                        end_iter = first_iter
                    else:
                        end_iter = self.get_len() - 1

                blp_next_off = self.get_blinepoint(end_iter, off_time)

                begin_iter = iterr
                blp_prev_off.set_origin(100)

                while True:
                    if begin_iter == 0:
                        if self.get_loop():
                            begin_iter = self.get_len()
                        else:
                            break
                    begin_iter -= 1
                    dist_from_begin += 1

                    if begin_iter == iterr:
                        break

                    if self.get_entry_list()[begin_iter]["ActivepointList"].amount_at_time(fr) > amount:
                        blp_prev_off = self.get_blinepoint(begin_iter, off_time)
                        break

                if blp_prev_off.get_origin() == 100:
                    if first_flag:
                        begin_iter = 0
                    else:
                        begin_iter = first_iter
                    blp_prev_off = self.get_blinepoint(begin_iter, off_time)

                curve = Hermite(blp_prev_off.get_vertex(),
                                blp_next_off.get_vertex(),
                                blp_prev_off.get_tangent2(),
                                blp_next_off.get_tangent1())

                blp_here_off.set_vertex(curve.value(blp_here_on.get_origin()))
                blp_here_off.set_width((blp_next_off.get_width()-blp_prev_off.get_width())*blp_here_on.get_origin()+blp_prev_off.get_width())
                blp_here_off.set_tangent1(curve.derivative(blp_here_on.get_origin()));
                blp_here_off.set_tangent2(curve.derivative(blp_here_on.get_origin()));

                if begin_iter == (iterr - 1) or dist_from_begin == 1:
                    prev_tangent_scalar = self.linear_interpolation(blp_here_on.get_origin(), 1.0, amount)
                else:
                    prev_tangent_scalar = self.linear_interpolation(blp_here_on.get_origin()-prev.get_origin(), 1.0, amount)

                if end_iter == (iterr + 1) or dist_from_end == 1:
                    next_tangent_scalar= self.linear_interpolation(1.0-blp_here_on.get_origin(), 1.0, amount)
                elif self.get_len() != (iterr + 1):
                    nextt = self.get_blinepoint(iterr+1, fr)
                    next_tangent_scalar = self.linear_interpolation(nextt.get_origin()-blp_here_on.get_origin(), 1.0, amount)
                else:
                    next_tangent_scalar = self.linear_interpolation(blp_next_off.get_origin()-blp_here_on.get_origin(), 1.0, amount)
                next_scale=next_tangent_scalar

                # My second try
                off_coord_sys = []
                on_coord_sys = []
                curr_coord_sys = []
                end_pos_at_off_time = self.get_blinepoint(end_iter, off_time).get_vertex()
                begin_pos_at_off_time = self.get_blinepoint(begin_iter, off_time).get_vertex()
                off_coord_origin = (begin_pos_at_off_time + end_pos_at_off_time)/2
                off_coord_sys.append((begin_pos_at_off_time - end_pos_at_off_time).norm())
                off_coord_sys.append(off_coord_sys[0].perp())

                end_pos_at_on_time = self.get_blinepoint(end_iter, on_time).get_vertex()
                begin_pos_at_on_time = self.get_blinepoint(begin_iter, on_time).get_vertex()
                on_coord_origin = (begin_pos_at_on_time + end_pos_at_on_time)/2
                on_coord_sys.append((begin_pos_at_on_time - end_pos_at_on_time).norm())
                on_coord_sys.append(on_coord_sys[0].perp())

                end_pos_at_current_time = self.get_blinepoint(end_iter, fr).get_vertex()
                begin_pos_at_current_time = self.get_blinepoint(begin_iter, fr).get_vertex()
                curr_coord_origin = (begin_pos_at_current_time + end_pos_at_current_time)/2
                curr_coord_sys.append((begin_pos_at_current_time - end_pos_at_current_time).norm())
                curr_coord_sys.append(curr_coord_sys[0].perp())

                # swapping
                temp = curr_coord_sys[0][1]
                curr_coord_sys[0][1] = curr_coord_sys[1][0]
                curr_coord_sys[1][0] = temp

                trans_on_point = Vector(0, 0)
                trans_off_point = Vector(0, 0)
                trans_on_t1 = Vector(0, 0)
                trans_off_t1 = Vector(0, 0)
                trans_on_t2 = Vector(0, 0)
                trans_off_t2 = Vector(0, 0)

                trans_on_point = self.transform_coords(blp_here_on.get_vertex(),  trans_on_point,  on_coord_origin,  on_coord_sys)
                trans_off_point = self.transform_coords(blp_here_off.get_vertex(), trans_off_point, off_coord_origin, off_coord_sys)

                trans_on_t1 = self.transform_coords(blp_here_on.get_tangent1(),  trans_on_t1,  Vector(0, 0), on_coord_sys)
                trans_off_t1 = self.transform_coords(blp_here_off.get_tangent1(), trans_off_t1, Vector(0, 0), off_coord_sys);

                if blp_here_on.get_split_tangent_both():
                    trans_on_t2 = self.transform_coords(blp_here_on.get_tangent2(),  trans_on_t2,  Vector(0, 0), on_coord_sys)
                    trans_off_t2 = self.transform_coords(blp_here_off.get_tangent2(), trans_off_t2, Vector(0, 0), off_coord_sys)

                tmp = Vector(0, 0)
                tmp = self.untransform_coords(self.linear_interpolation(trans_off_point, trans_on_point, amount), tmp, curr_coord_origin, curr_coord_sys)
                blp_here_now.set_vertex(tmp)

                tmp = Vector(0, 0)
                tmp = self.untransform_coords(self.radial_interpolation(trans_off_t1,trans_on_t1,amount), tmp, Vector(0, 0), curr_coord_sys)
                blp_here_now.set_tangent1(tmp)

                # blp_here_now.set_tangent1(self.radial_interpolation(blp_here_off.get_tangent1(), blp_here_on.get_tangent1(), amount))

                if blp_here_on.get_split_tangent_both():
                    blp_here_now.set_split_tangent_both(True)
                    tmp = Vector(0, 0)
                    tmp = self.untransform_coords(self.radial_interpolation(trans_off_t2,trans_on_t2,amount), tmp, Vector(0, 0), curr_coord_sys)
                    blp_here_now.set_tangent2(tmp)
                else:
                    blp_here_now.set_split_tangent_both(False)

                blp_here_now.set_origin(blp_here_on.get_origin())
                blp_here_now.set_width(self.linear_interpolation(blp_here_off.get_width(), blp_here_on.get_width(), amount))

                if first_flag:
                    blp_here_now.set_tangent1(blp_here_now.get_tangent1()*prev_tangent_scalar)
                    first_iter = iterr
                    first = prev = blp_here_now
                    first_flag = False
                    ret_list.append(copy.deepcopy(blp_here_now))
                    iterr += 1
                    continue

                ret_list[-1].set_split_tangent_both(True)
                ret_list[-1].set_tangent2(prev.get_tangent2()*prev_tangent_scalar)

                ret_list.append(copy.deepcopy(blp_here_now))
                ret_list[-1].set_split_tangent_both(True)
                ret_list[-1].set_tangent1(blp_here_now.get_tangent1()*prev_tangent_scalar)
                prev = blp_here_now

            iterr += 1

        if next_scale != 1:
            ret_list[-1].set_split_tangent_both(True)
            ret_list[-1].set_tangent2(prev.get_tangent2()*next_scale)

        return ret_list

    def radial_interpolation(self, a, b, c):
        """
        """
        if a.is_equal_to(Vector(0, 0)) or b.is_equal_to(Vector(0, 0)):
            return linear_interpolation(a, b, c)

        mag = self.mag_combo(a.mag(), b.mag(), c)
        angle_a = TanAngle(Vector(a[1], a[0]))
        angle_b = TanAngle(Vector(b[1], b[0]))
        diff = DegAngle(angle_b - angle_a).get()
        if diff < -180: angle_b += DegAngle(360)
        elif diff > 180:    angle_a += DegAngle(360)
        ang = self.ang_combo(angle_a, angle_b, c)

        return Vector(mag*CosAngle(ang).get(), mag*SinAngle(ang).get())

    def mag_combo(self, a, b, t):
        return (b-a)*t+a

    def ang_combo(self, a, b, t):
        return b.dist(a)*t + a

    def untransform_coords(self, inn, out, coord_origin, coord_sys):
        out[0] = inn * coord_sys[0] 
        out[1] = inn * coord_sys[1]
        out += coord_origin
        return out

    def transform_coords(self, inn, out, coord_origin, coord_sys):
        inn -= coord_origin
        out[0] = inn * coord_sys[0]
        out[1] = inn * coord_sys[1]
        return out

    def linear_interpolation(self, a, b, c):
        return (b-a)*c + a

    def get_blinepoint(self, current, t):
        """
        """
        bpcurr = self.get_value_at_frame(self.get_entry_list()[current], t)
        if not bpcurr.get_boned_vertex_flag():
            return bpcurr

        nxt = current
        previous = current
        
        nxt += 1
        if nxt == self.get_len():
            nxt = 0
        if current == 0:
            previous = self.get_len()
        previous -= 1

        bpprev = self.get_value_at_frame(self.get_entry_list()[previous], t)
        bpnext = self.get_value_at_frame(self.get_entry_list()[nxt], t)

        t1 = bpcurr.get_tangent1()
        t2 = bpcurr.get_tangent2()
        v = bpcurr.get_vertex()
        vp = bpprev.get_vertex()
        vn = bpnext.get_vertex()
        vs = bpcurr.get_vertex_setup()
        vps = bpprev.get_vertex_setup()
        vns = bpnext.get_vertex_setup()
        beta01 = t1.angle()
        beta02 = t2.angle()

        alpha = (vn - vp).angle() - (vns - vps).angle()
        if bpcurr.get_split_tangent_both():
            gamma = ((v-(vn+vp)*0.5).angle()-(vn-vp).angle()) - ((vs-(vns+vps)*0.5).angle()-(vns-vps).angle())
        else:
            gamma = Angle(0)

        beta1 = alpha + gamma + beta01
        beta2 = alpha + gamma + beta02

        tt1 = Vector(t1.mag()*CosAngle(beta1).get(), t1.mag()*SinAngle(beta1).get())
        tt2 = Vector(t2.mag()*CosAngle(beta2).get(), t2.mag()*SinAngle(beta2).get())

        bpcurr.set_tangent1(tt1)
        bpcurr.set_tangent2(tt2)

        return bpcurr




"""
ActivepointList.py
To store the on/off times of spline points

REFER HERE:
https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/activepoint.h
"""

import sys
import settings
from common.Activepoint import Activepoint
sys.path.append("..")


class ActivepointList:
    """
    Class to store the active points of Synfig
    """
    def __init__(self, on_time, off_time):
        self.active_point_list = []
        if on_time is None and off_time is None:
            return

        if on_time is not None:
            self.on_time = on_time.split(",")
            for time in self.on_time:
                self.active_point_list.append(Activepoint(time, True))

        if off_time is not None:
            self.off_time = off_time.split(",")
            for time in self.off_time:
                self.active_point_list.append(Activepoint(time, False))

        # Sort this list, and remove duplicates if any
        # Removing duplicates
        self.active_point_list = self.remove_duplicates()
        self.active_point_list = sorted(self.active_point_list)

    def remove_duplicates(self):
        """
        https://stackoverflow.com/questions/480214/how-do-you-remove-duplicates-from-a-list-whilst-preserving-order

        If at a single time we have two stops, then remove one of that
        """
        seq = self.active_point_list
        seen = set()
        seen_add = seen.add
        return [x for x in seq if not (x.time in seen or seen_add(x.time))]

    def empty(self):
        return len(self.active_point_list) == 0

    def find(self, frame):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_dynamiclist.cpp#L268
        """
        for itr in self.active_point_list:
            if itr.time == frame:
                return itr

        raise Exception("Value not found")

    def find_prev(self, frame):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_dynamiclist.cpp#L324
        """
        itr = len(self.active_point_list)

        while True:
            itr -= 1
            if self.active_point_list[itr].time < frame:
                return self.active_point_list[itr]

            if not (itr != 0):
                break

        raise Exception("Could not find prev")

    def find_next(self, frame):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_dynamiclist.cpp#L296
        """
        for itr in self.active_point_list:
            if itr.time > frame:
                return itr

        raise Exception("Could not find next")

    def amount_at_time(self, frame, rising = None):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_dynamiclist.cpp#L394
        """
        if self.empty():
            return 1

        try:
            itr = self.find(frame)
            return 1 if itr.state else 0
        except Exception as e:
            pass

        try:
            prev_itr = self.find_prev(frame)
        except Exception as e:
            return 1 if self.find_next(frame).state else 0

        try:
            next_itr = self.find_next(frame)
        except Exception as e:
            return 1 if prev_itr.state else 0

        if next_itr.state == prev_itr.state:
            return 1 if next_itr.state else 0 

        if rising is not None:
            rising[0] = next_itr.state

        if next_itr.state == True:
            return float((frame - prev_itr.time)/(next_itr.time - prev_itr.time))
        return float((next_itr.time - frame)/(next_itr.time - prev_itr.time))

    def status_at_time(self, t):
        """
        https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_dynamiclist.cpp#L441
        """
        state = True
        if not self.empty():
            if len(self.active_point_list) == 1:
                state = self.active_point_list[0].state
            else:
                entry_itr = 0
                while entry_itr != len(self.active_point_list):
                    if self.active_point_list[entry_itr].time == t:
                        return self.active_point_list[entry_itr].state
                    if self.active_point_list[entry_itr].time > t:
                        break
                    entry_itr += 1
                prev_itr = entry_itr
                prev_itr -= 1

                if entry_itr == len(self.active_point_list):
                    state = self.active_point_list[prev_itr].state
                elif entry_itr == 0:
                    state = self.active_point_list[entry_itr].state
                elif self.active_point_list[entry_itr].priority == self.active_point_list[prev_itr].priority:
                    state = self.active_point_list[entry_itr].state or self.active_point_list[prev_itr].state
                elif self.active_point_list[entry_itr].priority > self.active_point_list[prev_itr].priority:
                    state = self.active_point_list[entry_itr].state
                else:
                    state = self.active_point_list[prev_itr].state

        return state

    def update_frame_window(self, window):
        """
        Find the minimum and maximum points at which the waypoints are located
        """
        if self.empty():
            return

        for itr in self.active_point_list:
            fr = itr.time
            if fr == settings.SOT or fr == settings.EOT:
                continue

            if fr > window["last"]:
                window["last"] = fr
            if fr < window["first"]:
                window["first"] = fr



# pylint: disable=line-too-long
"""
Implements function required for generating tangent between two value key frames
"""

import sys
import random
import settings
from common.misc import get_frame, parse_position
from properties.offsetKeyframe import calc_tangent
sys.path.append("../")


def normalize_tangents(cur_pos, next_pos, t_in, t_out):
    """
    Converts the tangents from arbitrary range to the range of 0-1

    Args:
        cur_pos  (common.Vector.Vector) : Current position in coordinate system
        next_pos (common.Vector.Vector) : Next position in coordinate system
        t_in     (dict)        : In tangent in Lottie format
        t_out    (dict)        : Out tangent in Lottie format

    Returns:
        (None)
    """
    # time_scale means converting time(on x-axis) to 0-1 range
    time_scale = next_pos[1] - cur_pos[1]

    # value_scale -> converting value(on y-axis to 0-1 range)
    value_scale = next_pos[0] - cur_pos[0]

    # If ever the value scale equals to absolute zero, randomly add or subtract
    # 1e-9 to it, in order to avoid division by zero
    # This difference in value is not caught by bare human eyes, so should not effect
    if value_scale == 0.0:
        bias = 1 if random.random() < 0.5 else -1
        bias *= 1e-9
        value_scale += bias

    # Trying the same for time_scale
    if time_scale == 0.0:
        bias = 1 if random.random() < 0.5 else -1
        bias *= 1e-9
        time_scale += bias

    time_diff = cur_pos[1] / time_scale
    value_diff = cur_pos[0] / value_scale

    t_out["x"][0] += cur_pos[1]     # The tangents were relative to cur_pos, now relative to 0
    t_out["y"][0] += cur_pos[0]

    # -ve sign because lottie and synfig use opposite "in" tangents
    t_in["x"][0] = next_pos[1] - t_in["x"][0]
    t_in["y"][0] = next_pos[0] - t_in["y"][0]

    # abs is put in order to deal with decreasing value
    t_out["x"][0] = abs(t_out["x"][0] / time_scale - time_diff)
    t_out["y"][0] = abs(t_out["y"][0] / value_scale - value_diff)

    t_in["x"][0] = abs(t_in["x"][0] / time_scale - time_diff)
    t_in["y"][0] = abs(t_in["y"][0] / value_scale - value_diff)


def gen_value_Keyframe(curve_list, animated, i):
    """
    Generates the dictionary corresponding to properties/valueKeyframe.json in lottie
    documentation

    Args:
        curve_list (list)                : Bezier curve in Lottie format
        animated   (lxml.etree._Element) : Synfig format animation
        i          (int)                 : Iterator for animation

    Returns:
        (TypeError) : If hold interval is encountered
        (None)      : Otherwise
    """
    lottie = curve_list[-1]
    waypoint, next_waypoint = animated[i], animated[i+1]
    cur_get_after, next_get_before = waypoint.attrib["after"], next_waypoint.attrib["before"]
    cur_get_before, next_get_after = waypoint.attrib["before"], next_waypoint.attrib["after"]
    # Calculate positions of waypoints
    if animated.attrib["type"] in {"angle", "star_angle_new", "region_angle"}:
    #if animated.attrib["type"] in {"angle"}:
        if cur_get_after == "auto":
            cur_get_after = "linear"
        if cur_get_before == "auto":
            cur_get_before = "linear"
        if next_get_before == "auto":
            next_get_before = "linear"
        if next_get_after == "auto":
            next_get_after = "linear"

    # Synfig only supports constant interpolations for points
    if animated.attrib["type"] == "points":
        cur_get_after = "constant"
        cur_get_before = "constant"
        next_get_after = "constant"
        next_get_before = "constant"

    # After effects only supports linear,ease-in,ease-out and constant interpolations for color
    ##### No support for TCB and clamped interpolations in color is there yet #####
    if animated.attrib["type"] == {"color", "gradient"}:
        if cur_get_after in {"auto", "clamped"}:
            cur_get_after = "linear"
        if cur_get_before in {"auto", "clamped"}:
            cur_get_before = "linear"
        if next_get_before in {"auto", "clamped"}:
            next_get_before = "linear"
        if next_get_after in {"auto", "clamped"}:
            next_get_after = "linear"

    cur_pos = parse_position(animated, i)
    next_pos = parse_position(animated, i + 1)

    lottie["t"] = get_frame(waypoint)
    lottie["s"] = cur_pos.get_val()
    lottie["e"] = next_pos.get_val()

    lottie["i"] = {}
    lottie["o"] = {}

    try:
        out_val, in_val = calc_tangent(animated, lottie, i)
    except Exception as excep:
        # That means halt/constant interval
        return excep

    set_tangents(out_val, in_val, cur_pos, next_pos, lottie, animated)

    if cur_get_after == "halt": # For ease out
        lottie["o"]["x"][0] = settings.OUT_TANGENT_X
        lottie["o"]["y"][0] = settings.OUT_TANGENT_Y
        lottie["synfig_o"] = [0]
    if next_get_before == "halt": # For ease in
        lottie["i"]["x"][0] = settings.IN_TANGENT_X
        lottie["i"]["y"][0] = settings.IN_TANGENT_Y
        lottie["synfig_i"] = [0]

    # TCB/!TCB and list is not empty
    if cur_get_before == "auto" and cur_get_after != "auto" and i > 0:

        # need value for previous tangents
        # It may be helpful to store them somewhere
        prev_ov, prev_iv = calc_tangent(animated, curve_list[-2], i - 1)
        prev_iv = out_val
        set_tangents(prev_ov, prev_iv, parse_position(animated, i-1), cur_pos, curve_list[-2], animated)
        if cur_get_after == "halt":
            curve_list[-2]["i"]["x"][0] = settings.IN_TANGENT_X
            curve_list[-2]["i"]["y"][0] = settings.IN_TANGENT_Y
            lottie["synfig_i"] = [0]


def set_tangents(out_val, in_val, cur_pos, next_pos, lottie, animated):
    """
    To set the tangents as required by the lottie format for value waypoints

    Args:
        out_val   (common.Vector.Vector)         : Tangent out value
        in_val    (common.Vector.Vector)         : Tangent in value
        cur_pos   (common.Vector.Vector)         : Current position in coordinate system
        next_pos  (common.Vector.Vector)         : Next position in coordinate system
        lottie    (dict)                : bezier interval in lottie format
        animation (lxml.etree._Element) : Synfig format animation

    Returns:
        (None)
    """
    out_lst = out_val.get_list()
    in_lst = in_val.get_list()

    lottie["i"]["x"] = [in_lst[1]]    # x-axis represents time
    lottie["i"]["y"] = [in_lst[0]]    # y-axis represents animation progress
    lottie["o"]["x"] = [out_lst[1]]
    lottie["o"]["y"] = [out_lst[0]]

    # Lottie tangent length is larger than synfig
    lottie["i"]["x"][0] /= settings.TANGENT_FACTOR
    lottie["i"]["y"][0] /= settings.TANGENT_FACTOR
    lottie["o"]["x"][0] /= settings.TANGENT_FACTOR
    lottie["o"]["y"][0] /= settings.TANGENT_FACTOR

    # Keeping original Synfig format tangents in dictionary
    lottie["synfig_i"] = [lottie["i"]["y"][0]]
    lottie["synfig_o"] = [lottie["o"]["y"][0]]

    # If type is color, the tangents are already normalized
    if animated.attrib["type"] not in {"color", "gradient"}:
        normalize_tangents(cur_pos, next_pos, lottie["i"], lottie["o"])



# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of polygon layer
in Lottie format
"""

import sys
from common.Vector import Vector
from common.Bline import Bline
from common.Param import Param
from properties.shapePropKeyframe.helper import insert_dict_at
sys.path.append("../../")


def gen_dynamic_list_polygon(lottie, dynamic_list):
    """
    Generates the bline corresponding to polygon layer

    Args:
        lottie (dict) : Lottie format polygon layer will be stored here
        dynamic_list (common.Param.Param) : Synfig format points of polygon

    Returns:
        (None)
    """
    ################## SECTION 1 ################
    # Inserting the waypoints if not animated, finding the first and last frame
    # Calculating the path after this
    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1
    dynamic_list = Bline(dynamic_list[0], dynamic_list)

    for entry in dynamic_list.get_entry_list():
        pos = entry["vector"]
        pos.update_frame_window(window)

        z = Param(pos.getparent(), pos.getparent().getparent())
        z.animate("vector", True)
        entry["vector"] = z

    layer = dynamic_list.get_layer().get_layer()
    origin = layer.get_param("origin")

    # Animating the origin
    origin.update_frame_window(window)
    origin.animate("vector")

    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    ################ END OF SECTION 1 ##############

    ################ SECTION 2 #####################
    # Generating values for all the frames in the window
    fr = window["first"]
    while fr <= window["last"]:
        st_val, en_val = insert_dict_at(lottie, -1, fr, False)

        for entry in dynamic_list.get_entry_list():
            pos_cur = entry["vector"].get_value(fr)
            pos_next = entry["vector"].get_value(fr + 1)

            tangent1_cur, tangent2_cur = Vector(0, 0), Vector(0, 0)
            tangent1_next, tangent2_next = Vector(0, 0), Vector(0, 0)

            # Adding origin to each vertex
            origin_cur = origin.get_value(fr)
            origin_next = origin.get_value(fr + 1)
            for i in range(len(pos_cur)):
                pos_cur[i] += origin_cur[i]
            for i in range(len(pos_next)):
                pos_next[i] += origin_next[i]

            # Store values in dictionary
            st_val["i"].append(tangent1_cur.get_list())
            st_val["o"].append(tangent2_cur.get_list())
            st_val["v"].append(pos_cur)
            en_val["i"].append(tangent1_next.get_list())
            en_val["o"].append(tangent2_next.get_list())
            en_val["v"].append(pos_next)
        fr += 1
    # Setting the final time
    lottie.append({})
    lottie[-1]["t"] = fr



"""
Stores all the functions required for generating value key frames in lottie
"""

import sys
from properties.timeAdjust import time_adjust
from properties.valueKeyframe import gen_value_Keyframe
from common.misc import get_frame
sys.path.append("../")


def gen_value_Keyframed(lottie, animated, idx):
    """
    Generates the dictionary corresponding to properties/valueKeyframed.json in
    lottie documentation

    Args:
        lottie (dict)                  : Lottie bezier curve stored in this
        animated (lxml.etree._Element) : Synfig format animation
        idx      (int)                 : Index of animation

    Returns:
        (None)
    """
    lottie["ix"] = idx
    lottie["a"] = 1
    lottie["k"] = []
    for i in range(len(animated) - 1):
        lottie["k"].append({})
        gen_value_Keyframe(lottie["k"], animated, i)
    last_waypoint_frame = get_frame(animated[-1])
    lottie["k"].append({})
    lottie["k"][-1]["t"] = last_waypoint_frame

    if "h" in lottie["k"][-2].keys():
        lottie["k"][-1]["h"] = 1
        lottie["k"][-1]["s"] = lottie["k"][-2]["e"]

        # specific case for points when prev_points > cur_points
        if animated.attrib["type"] == "points":
            if lottie["k"][-2]["s"][0] > lottie["k"][-1]["s"][0]:
                # Adding 1 frame to the previous time
                prev_frames = get_frame(animated[-2])
                lottie["k"][-1]["t"] = prev_frames + 1

    time_adjust(lottie, animated)



"""
Will store all functions needed to generate the shape(path) layer in lottie
# This is not currently used anywhere because we are using solid_shape.py instead of this module to
support the invert parameter
"""

import sys
from common.Count import Count
from properties.shapeKeyframed import gen_properties_shapeKeyframed
sys.path.append("..")


def gen_shapes_shape(lottie, layer, idx):
    """
    Generates the dictionary corresponding to shapes/shape.json

    Args:
        lottie (dict)               : The lottie generated shape layer will be stored in it
        layer  (lxml.etree._Element): Synfig format shape(list of bline points) layer
        idx    (int)                : Stores the index of the shape layer

    Returns:
        (None)
    """
    index = Count()
    lottie["ty"] = "sh"     # Type: shape
    lottie["ix"] = idx      # Index
    lottie["ks"] = {}
    gen_properties_shapeKeyframed(lottie["ks"], layer, index.inc())



# pylint: disable=line-too-long
"""
Will store all the functions and modules for generation of Advanced outline layer
in Lottie format
"""

import sys
import math
import copy
from common.Bline import Bline
from common.WidthPoint import WidthPoint
from common.WidthPointList import WidthPointList
from common.DashItemList import DashItemList
from common.Vector import Vector
from common.Hermite import Hermite
from common.Angle import RadAngle, SinAngle, CosAngle
from synfig.animation import to_Synfig_axis
from properties.shapePropKeyframe.helper import add, insert_dict_at_adv_outline, animate_tangents
from properties.shapePropKeyframe.outline import line_intersection, get_outline_grow
sys.path.append("../../")


def gen_bline_advanced_outline(lottie, bline_point):
    """
    Generates the bline corresponding to advanced outline layer by adding some vertices
    to bline and converting it to region layer

    Args:
        lottie (dict) : Lottie format outline layer will be stored in this
        bline_point (common.Param.Param) : Synfig format bline points

    Returns:
        (None)
    """
    ################### SECTION 1 #########################
    # Inserting waypoints if not animated and finding the first and last frame
    window = {}
    window["first"] = sys.maxsize
    window["last"] = -1

    bline = Bline(bline_point[0], bline_point)

    for entry in bline.get_entry_list():
        pos = entry["point"]
        origin = entry["origin"]
        width = entry["width"]
        t1 = entry["t1"]
        t2 = entry["t2"]
        split_r = entry["split_radius"]
        split_a = entry["split_angle"]

        pos.update_frame_window(window)
        # Empty the pos and fill in the new animated pos
        pos.animate("vector")

        origin.update_frame_window(window)
        origin.animate("real")

        width.update_frame_window(window)
        width.animate("real")

        split_r.update_frame_window(window)
        split_r.animate_without_path("bool")

        split_a.update_frame_window(window)
        split_a.animate_without_path("bool")

        animate_tangents(t1, window)
        animate_tangents(t2, window)

        entry["ActivepointList"].update_frame_window(window)

    layer = bline.get_layer().get_layer()

    # For width point list
    width_point_list = layer.get_param("wplist")
    width_point_list = WidthPointList(width_point_list[0], width_point_list, bline.get_loop())

    for entry in width_point_list.get_entry_list():
        pos = entry["position"]
        width = entry["width"]
        # All other remaining parameters are static

        pos.update_frame_window(window)
        pos.animate("real")

        width.update_frame_window(window)
        width.animate("real")

        entry["ActivepointList"].update_frame_window(window)

    # For Dash Item list
    dash_item_list = layer.get_param("dilist")
    dash_item_list = DashItemList(dash_item_list[0], dash_item_list)

    for entry in dash_item_list.get_entry_list():
        offset = entry["offset"]
        length = entry["length"]
        # All other remaining parameters are static

        offset.update_frame_window(window)
        offset.animate("real")

        length.update_frame_window(window)
        length.animate("real")

        # Update frame window if any list point is added/deleted
        entry["ActivepointList"].update_frame_window(window)

    outer_width = layer.get_param("width")
    expand = layer.get_param("expand")
    
    start_tip = layer.get_param("start_tip").get()  # Static parameter
    start_tip = int(start_tip[0].attrib["value"])
    end_tip = layer.get_param("end_tip").get()      # Static parameter: can't be animated
    end_tip = int(end_tip[0].attrib["value"])
    cusp_type = layer.get_param("cusp_type").get()  # Static parameter
    cusp_type = int(cusp_type[0].attrib["value"])
    homogeneous = layer.get_param("homogeneous")    # Static parameter
    homogeneous = True if homogeneous[0].attrib["value"] == "true" else False

    smoothness = layer.get_param("smoothness")
    dash_enabled = layer.get_param("dash_enabled")
    dash_offset = layer.get_param("dash_offset")
    origin = layer.get_param("origin")

    # Animating the origin
    origin.update_frame_window(window)
    origin.animate("vector")

    # Animating the outer width
    outer_width.update_frame_window(window)
    outer_width.animate("real")

    # Animating smoothness
    smoothness.update_frame_window(window)
    smoothness.animate("real")

    # Animating the expand param
    expand.update_frame_window(window)
    expand.animate("real")

    # Animating the dash enabled
    dash_enabled.update_frame_window(window)
    dash_enabled.animate_without_path("bool")

    # Animating the dash offset
    dash_offset.update_frame_window(window)
    dash_offset.animate("real")

    # Minimizing the window size
    if window["first"] == sys.maxsize and window["last"] == -1:
        window["first"] = window["last"] = 0
    ################# END OF SECTION 1 ###################

    ################ SECTION 2 ###########################
    # Generating values for all the frames in the window

    # Store all side_a, side_b values; because we need to make them equal in
    # size in order to render properly in lottie
    st_list = []
    en_list = []
    lottie_st = []
    lottie_en = []

    fr = window["first"]
    while fr <= window["last"]:
        st_val, en_val = insert_dict_at_adv_outline(lottie, -1, fr, False)  # This loop needs to be considered somewhere down
        lottie_st.append([st_val, fr])
        lottie_en.append([en_val, fr+1])

        st_list_value = synfig_advanced_outline(bline, outer_width, expand,
                start_tip, end_tip, cusp_type, smoothness, homogeneous,
                dash_enabled, dash_offset, dash_item_list, width_point_list,
                fr)
        en_list_value = synfig_advanced_outline(bline, outer_width, expand,
                start_tip, end_tip, cusp_type, smoothness, homogeneous,
                dash_enabled, dash_offset, dash_item_list, width_point_list,
                fr + 1)
        st_list.append(st_list_value)
        en_list.append(en_list_value)

        fr += 1

    append_all_lists(st_list, en_list, lottie_st, lottie_en, origin)

    # Setting the final time
    lottie.append({})
    lottie[-1]["t"] = fr
    lottie[-1]["h"] = 1
    # Need to define last frame's "s" value if we have hold interpolation
    lottie[-1]["s"] = lottie[-2]["s"]


def append_all_lists(st_list, en_list, lottie_st, lottie_en, origin_p):
    """
    This function collects all the lists(at each different frame), then
    makes all those lists of equal size(lottie style), and then appends them to
    the lottie dictionary

    Args:
        st_list (list[Floats]): These are the points obtained from synfig at frame "st"
        en_list (list[Floats]): These are the points obtained from synfig at frame "en"
        lottie_en (list[(dict, Int)]): Contains address of lottie's "en" dict and frame's value
        origin_p (common.param.param): Lottie format origin of adv outline layer
    
    Returns:
        (None)
    """
    # Find maximum size among all the lists
    mx = 0
    for st in st_list:
        mx = max(mx, len(st))
    for en in en_list:
        mx = max(mx, len(en))

    # All the lists should of size mx
    for st in st_list:
        diff = mx - len(st)
        last_value = copy.deepcopy(st[-1])
        st.extend([last_value for i in range(diff)])

    for en in en_list:
        diff = mx - len(en)
        last_value = copy.deepcopy(en[-1])
        en.extend([last_value for i in range(diff)])

    # Now render 
    for i in range(0, len(lottie_st)):
        cur_frame = lottie_st[i][1]
        origin_cur = origin_p.get_value(cur_frame)
        add(st_list[i], lottie_st[i][0], origin_cur) 
        lottie_st[i][0]["h"] = 1

    for i in range(0, len(lottie_en)):
        cur_frame = lottie_en[i][1]
        origin_cur = origin_p.get_value(cur_frame)
        add(en_list[i], lottie_en[i][0], origin_cur)
        lottie_en[i][0]["h"] = 1


def synfig_advanced_outline(bline, outer_width_p, expand_p, start_tip, end_tip,
        cusp_type, smoothness_p, homogeneous, dash_enabled_p, dash_offset_p,
        dash_item_list, width_point_list, fr):
    """
    Calculates the points for the advanced outline layer as in Synfig:
    https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/modules/mod_geometry/advanced_outline.cpp

    Args:
        bline (common.Bline.Bline): Synfig format bline points of adv outline layer
        outer_width_p (common.Param.Param): Lottie format outer width
        expand_p (common.Param.Param): Lottie format expand parameter
        start_tip (Int): Static value of start tip of adv outline
        end_tip (Int): Static value of end tip of adv outline
        cusp_type (Int): Static value of cusp type
        smoothness_p (common.Param.Param): Lottie format smoothness parameter
        homogeneous (Bool): Static value of homogeneous parameter
        dash_enabled_p (common.Param.Param): Lottie format dash_enabled parameter
        dash_offset_p (common.Param.Param): Lottie format dash_offset parameter
        dash_item_list (common.DashItemList.DashItemList): Dash item list of Synfig
        width_point_list (common.WidthPointList.WidthPointList): Width point list of Synfig

    Returns:
        (list[(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector)]): (2D points value, tangent1, tangent2)
    """
    CUSP_TANGENT_ADJUST = 0.025
    SAMPLES = 50
    EPSILON = 0.000000001
    fast_ = False   # This parameter has been removed from Synfig 1.4 onwards
    smoothness = to_Synfig_axis(smoothness_p.get_value(fr), "real")
    dash_offset = to_Synfig_axis(dash_offset_p.get_value(fr), "real")
    dash_enabled = dash_enabled_p.get_value(fr)
    expand = to_Synfig_axis(expand_p.get_value(fr), "real")
    width = to_Synfig_axis(outer_width_p.get_value(fr), "real")

    bline_list = bline.get_list_at_frame(fr)
    bline_pos, hbline_pos = [], []

    wplist = width_point_list.get_list_at_frame(fr)
    dilist = dash_item_list.get_list_at_frame(fr)
    dwplist = []
    fdwplist = []

    dash_enabled = dash_enabled and (dash_item_list.get_len() != 0)
    dstart_tip = 4  # WidthPointList::TYPE_FLAT
    dend_tip   = 4  # WidthPointList::TYPE_FLAT

    # Check if bline is looped
    blineloop = bline.get_loop()

    # Check if width point list is looped
    wplistloop = width_point_list.get_loop()

    bline_size = len(bline_list)
    wplist_size = width_point_list.get_len()

    # Initializing the iterator
    biter, bnext = 0, 0
    bpiter, bpnext, hbpiter = 0, 0, 0
    #witer, wnext, switer, swnext, cwiter, cwnext, scwiter, scwnext, dwiter, dwnext = 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

    first_tangent = None
    first = True
    middle_corner = False
    done_tip = False
    inserted_first = False
    inserted_last = False

    bezier_size = 1.0 / (bline_size if blineloop else (1.0 if bline_size==1 else (bline_size-1)))

    # Iterator
    bend = len(bline_list)

    # Outline grow
    gv = get_outline_grow(fr)

    if not blineloop and bline_size == 1:
        return

    bindex = 0.0
    biter = 0
    while biter != bend:
        bline_pos.append(bindex * bezier_size)
        hbline_pos.append(bline_pos[-1] if fast_ else std_to_hom(bline_list, bline_pos[-1], wplistloop, blineloop, fr))
        bindex += 1
        biter += 1

    if blineloop:
        bline_pos.append(1.0)
        hbline_pos.append(1.0)
    else:
        bline_pos.pop()
        bline_pos.append(1.0)
        hbline_pos.pop()
        hbline_pos.append(1.0)

    hbpiter = 0
    bpiter = 0
    biter_pos = bline_pos[bpiter]
    bpiter += 1
    hbpiter += 1
    bnext_pos = bline_pos[bpiter]
    hbnext_pos = hbline_pos[hbpiter]

    # Setup chunk list
    side_a, side_b = [], []
    
    for witer in wplist:
        witer.set_position(witer.get_norm_position(wplistloop))

    wplist = sorted(wplist)

    if blineloop:
        biter = len(bline_list) - 1
    else:
        biter = bnext
        bnext = bnext + 1

    last_tangent = bline_list[biter].get_tangent1()

    if blineloop and cusp_type == 0 and last_tangent.is_equal_to(Vector(0, 0)):
        curve = Hermite(bline_list[biter-1].get_vertex(),
                        bline_list[biter].get_vertex(),
                        bline_list[biter-1].get_tangent2(),
                        bline_list[biter].get_tangent1())
        last_tangent = curve.derivative(1.0 - CUSP_TANGENT_ADJUST)

    if not blineloop:
        if wplist_size != 0:
            wpfront = wplist[0]
            wpback = wplist[-1]
            if wpfront.get_side_type_before() == 0 and wpfront.get_position() != 0.0:
                wplist.append(WidthPoint(0.0, wpfront.get_width(), start_tip, 0))
            if wpback.get_side_type_after() == 0 and wpback.get_position() != 1.0:
                wplist.append(WidthPoint(1.0, wpback.get_width(), 0, end_tip))
        else:
            wplist.append(WidthPoint(0.0, 1.0, start_tip, 0))
            wplist.append(WidthPoint(1.0, 1.0, 0, end_tip))
    else:
        if wplist_size != 0:
            wpfront = wplist[0]
            wpback = wplist[-1]
            wpfb_int = (wpfront.get_side_type_before() == 0)
            wpba_int = (wpback.get_side_type_after() == 0)

            if wpfb_int or wpba_int:
                if wpfront.get_position() != 0.0:
                    i = copy.deepcopy(wpback)
                    n = copy.deepcopy(wpfront)
                    if not homogeneous and not fast_:
                        i.set_position(std_to_hom(bline_list, i.get_position(), wplistloop, blineloop, fr)) 
                        n.set_position(std_to_hom(bline_list, n.get_position(), wplistloop, blineloop, fr)) 
                    wplist.append(WidthPoint(0.0, widthpoint_interpolate(i, n, 0.0, smoothness), 0, 0))
                    inserted_first = True
                if wpback.get_position() != 1.0:
                    i = copy.deepcopy(wpback)
                    n = copy.deepcopy(wpfront)
                    if not homogeneous and not fast_:
                        i.set_position(std_to_hom(bline_list, i.get_position(), wplistloop, blineloop, fr)) 
                        n.set_position(std_to_hom(bline_list, n.get_position(), wplistloop, blineloop, fr)) 
                    wplist.append(WidthPoint(1.0, widthpoint_interpolate(i, n, 1.0, smoothness), 0, 0))
                    inserted_last = True
        else:
            wplist.append(WidthPoint(0.0, 1.0, 0, 0))
            wplist.append(WidthPoint(1.0, 1.0, 0, 0))

    wplist = sorted(wplist)

    step = 1.0/SAMPLES/bline_size

    if dash_enabled:
        blinelength = bline_length(bline_list, blineloop, None, fr)
        if blinelength > EPSILON:
            dashes_length = 0.0
            diter = 0
            rditer = len(dilist) - 1 
            before, after = None, None  # These are WidthPoints
            while diter != len(dilist):
                dashes_length += dilist[diter].get_length() + dilist[diter].get_offset()
                diter += 1
            if dashes_length > EPSILON:
                if math.fabs(dash_offset) > dashes_length:
                    dash_offset = math.fmod(dash_offset, dashes_length)
                dpos = dash_offset if dash_offset >= 0 else (dashes_length + dash_offset)
                diter = 0
                inserted_to_blinelength = 0
                while dpos < blinelength:
                    before_pos = (dpos + dilist[diter].get_offset())/blinelength
                    after_pos = (dpos + dilist[diter].get_offset() + dilist[diter].get_length())/blinelength 
                    before_pos = before_pos if homogeneous else hom_to_std(bline_list, before_pos, wplistloop, blineloop, fr)
                    after_pos = after_pos if homogeneous else hom_to_std(bline_list, after_pos, wplistloop, blineloop, fr)
                    before = WidthPoint(before_pos, 1.0, dilist[diter].get_side_type_before(), 0, True)
                    after = WidthPoint(after_pos, 1.0, 0, dilist[diter].get_side_type_after(), True)
                    dwplist.append(before)
                    dwplist.append(after)
                    dpos += dilist[diter].get_offset() + dilist[diter].get_length()
                    diter += 1
                    inserted_to_blinelength += 1
                    if diter == len(dilist):
                        diter = 0
                if inserted_to_blinelength != 0:
                    after = dwplist[-1]
                    if after.get_position() >= 1.0:
                        after.set_position(1.0)
                        dwplist.pop()
                        before = dwplist[-1]

                        if before.get_position() >= 1.0:
                            dwplist.pop()
                            inserted_to_blinelength -= 1
                        else:
                            dend_tip = after.get_side_type_after()
                            dwplist.append(after) 
                inserted_to_zero = 0

                dpos = dash_offset if (dash_offset >= 0) else (dashes_length + dash_offset)
                while dpos > 0.0:
                    before_pos = (dpos - dilist[rditer].get_length())/blinelength
                    after_pos = (dpos) / blinelength
                    before_pos = before_pos if homogeneous else hom_to_std(bline_list, before_pos, wplistloop, blineloop, fr)
                    after_pos = after_pos if homogeneous else hom_to_std(bline_list, after_pos, wplistloop, blineloop, fr)
                    before = WidthPoint(before_pos, 1.0, dilist[rditer].get_side_type_before(), 0, True)
                    after = WidthPoint(after_pos, 1.0, 0, dilist[rditer].get_side_type_after(), True)
                    dwplist.insert(0, after)
                    dwplist.insert(0, before)
                    dpos -= dilist[rditer].get_offset() + dilist[rditer].get_length()
                    rditer -= 1
                    inserted_to_zero += 1
                    if rditer == -1:
                        rditer = len(dilist) - 1

                if inserted_to_zero != 0:
                    before = dwplist[0]
                    if before.get_position() <= 0.0:
                        before.set_position(0.0)
                        dwplist.pop(0)
                        after = dwplist[0]

                        if after.get_position() <= 0.0:
                            dwplist.pop(0)
                            inserted_to_zero -= 1
                        else:
                            dstart_tip = before.get_side_type_before()
                            dwplist.insert(0, before)

                if inserted_to_blinelength == 0 and inserted_to_zero == 0:
                    before = WidthPoint(0.5, 1.0, 4, 0, True)
                    after = WidthPoint(0.5, 1.0, 0, 4, True)
                    dwplist.append(before)
                    dwplist.append(after)
                     
                wnext = 0
                if blineloop:
                    witer = len(wplist) - 1
                else:
                    witer = wnext

                while True: # Do while loop
                    witer_pos = wplist[witer].get_position()
                    wnext_pos = wplist[wnext].get_position()

                    if wplist[witer].get_side_type_after() == 0 or wplist[wnext].get_side_type_before() == 0:
                        dwiter = 0
                        while dwiter != len(dwplist):
                            dwiter_pos = dwplist[dwiter].get_position()
                            if dwiter_pos > witer_pos and dwiter_pos < wnext_pos:
                                fdwplist.append(copy.deepcopy(dwplist[dwiter]))
                            dwiter += 1
                    witer = wnext
                    wnext += 1

                    if not (wnext != len(wplist)):
                        break

                dwiter = 0
                dwnext = dwiter + 1
                
                while True: # Do while loop
                    dwiter_pos = dwplist[dwiter].get_position()
                    dwnext_pos = dwplist[dwnext].get_position()
                    witer = 0
                    while witer != len(wplist):
                        witer_pos = wplist[witer].get_position()
                        if witer_pos <= dwnext_pos and witer_pos >= dwiter_pos:
                            fdwplist.append(copy.deepcopy(wplist[witer]))
                        witer += 1
                    dwnext += 1
                    dwiter = dwnext
                    if dwnext == len(dwplist):
                        break
                    dwnext += 1

    cwplist = []
    for my_iterator in wplist:
        cwplist.append(copy.deepcopy(my_iterator))
    scwplist = []
    for my_iterator in wplist:
        scwplist.append(copy.deepcopy(my_iterator))

    if homogeneous:
        scwiter = 0
        while scwiter != len(scwplist):
            scwplist[scwiter].set_position(hom_to_std(bline_list, scwplist[scwiter].get_position(), wplistloop, blineloop, fr)) 
            scwiter += 1
    else:
        cwiter = 0
        while cwiter != len(cwplist):
            cwplist[cwiter].set_position(std_to_hom(bline_list, cwplist[cwiter].get_position(), wplistloop, blineloop, fr))
            cwiter += 1

    if dash_enabled:
        wplist = []
        for my_iterator in fdwplist:
            wplist.append(copy.deepcopy(my_iterator))
        wplist = sorted(wplist)
        witer = 0


    if len(wplist) == 0:
        wplist.append(WidthPoint(0.5, 1.0, 4, 4, True))

    swplist = []
    for my_iterator in wplist:
        swplist.append(copy.deepcopy(my_iterator))

    if homogeneous:
        switer = 0
        while switer != len(swplist):
            swplist[switer].set_position(hom_to_std(bline_list, swplist[switer].get_position(), wplistloop, blineloop, fr))
            switer += 1
    else:
        witer = 0
        while witer != len(wplist):
            wplist[witer].set_position(std_to_hom(bline_list, wplist[witer].get_position(), wplistloop, blineloop, fr))
            witer += 1
    
    wnext = 0
    swnext = 0
    if blineloop:
        witer = len(wplist) - 1
        #switer = len(swplist) - 1
    else:
        witer = wnext
        #switer = swnext
    cwnext = 0
    scwnext = 0
    if blineloop:
        cwiter = len(cwplist) - 1
        scwiter = len(scwplist) - 1
    else:
        cwiter = cwnext
        scwiter = scwnext
    wend = len(wplist)
    swend = len(wplist)

    ipos = 0.0
    hipos = 0.0

    if not blineloop:
        if wplist[wnext].get_position() == 0.0:
            wplist[wnext].set_side_type_before(dstart_tip if dash_enabled else start_tip)
        last = len(wplist) - 1
        if wplist[last].get_position() == 1.0:
            wplist[last].set_side_type_after(dend_tip if dash_enabled else end_tip)

    if dash_enabled:
        first = 0
        last = len(wplist) - 1
        if wplist[first].get_side_type_before() == 0:
            wplist[first].set_side_type_before(dstart_tip)
        if wplist[last].get_side_type_after() == 0:
            wplist[last].set_side_type_after(dend_tip)

    # Main loop
    while True:
        iter_t = bline_list[biter].get_tangent2()
        next_t = bline_list[bnext].get_tangent1()
        iter_t_mag = iter_t.mag()
        next_t_mag = next_t.mag()

        split_flag = bline_list[biter].get_split_tangent_angle() or (bline_list[biter].get_tangent1().mag() == 0) or (bline_list[biter].get_tangent2().mag() == 0)
        curve = Hermite(bline_list[biter].get_vertex(),
                        bline_list[bnext].get_vertex(),
                        iter_t,
                        next_t)
        if iter_t_mag == 0.0:
            iter_t = curve.derivative(CUSP_TANGENT_ADJUST)
        if next_t_mag == 0.0:
            next_t = curve.derivative(1.0 - CUSP_TANGENT_ADJUST) 
        if blineloop and first:
            first_tangent = iter_t
            first = False

        wnext_pos = wplist[wnext].get_position()
        swnext_pos = swplist[swnext].get_position()

        if ipos == swnext_pos:
            unitary = None
            hipos = wnext_pos
            bezier_ipos = bline_to_bezier(ipos, biter_pos, bezier_size)
            q = bezier_ipos
            if q < EPSILON:
                unitary = iter_t.norm()
            elif q > (1.0 - EPSILON):
                unitary = next_t.norm()
            else:
                unitary = curve.derivative(q).norm()
            if wplist[wnext].get_dash():
                ci = scwiter
                cn = scwnext
                if not fast_:    # fast is set to False always
                    ci = cwiter
                    cn = cwnext
                # ASSUMPTION: if fast is set to True, then the code will change below
                if cwplist[ci].get_position() == 0.0 and cwplist[cn].get_position() != 1.0 and inserted_first:
                    ci = len(cwplist) - 1
                    if inserted_last:
                        ci -= 1
                if cwplist[cn].get_position() == 1.0 and inserted_last:
                    cn = 0
                    if inserted_first:
                        cn += 1
                i = cwplist[ci]
                n = cwplist[cn]
                p = ipos
                if not fast_:
                    p = hipos
                wplist[wnext].set_width(widthpoint_interpolate(i, n, p, smoothness))
            add_tip(side_a, side_b, curve.value(q), unitary, wplist[wnext], gv, width, expand)

            witer = wnext
            wnext += 1
            swnext += 1

            if wnext == wend or swnext == swend:
                cwnext = 0
                cwiter = len(cwplist) - 1
                scwnext = 0
                scwiter = len(scwplist) - 1

                if blineloop and (bline_list[bnext].get_split_tangent_angle() or bline_list[bnext].get_tangent1().mag() == 0 or bline_list[bnext].get_tangent2().mag() == 0):
                    first = 0
                    last = len(wplist) - 1

                    if wplist[first].get_side_type_before() == 0 or wplist[last].get_side_type_after() == 0:
                        i = scwplist[scwiter]
                        n = scwplist[scwnext]
                        if not fast_:
                            i = cwplist[cwiter]
                            n = cwplist[cwnext]
                        p = ipos
                        if not fast_:
                            p = hipos
                        add_cusp(side_a,
                                 side_b,
                                 bline_list[bnext].get_vertex(),
                                 first_tangent,
                                 curve.derivative(1.0-CUSP_TANGENT_ADJUST),
                                 gv*(expand+width*0.5*widthpoint_interpolate(i, n, p, smoothness)),
                                 cusp_type)
                break
            else:
                ipos = ipos + EPSILON
                if wplist[witer].get_side_type_after() != 0:
                    done_tip = True
                else:
                    done_tip = False
                if ipos > scwplist[scwnext].get_position():
                    cwiter = cwnext
                    scwiter = scwnext
                    cwnext += 1
                    scwnext += 1
                middle_corner = False
                continue
        if (wplist[witer].get_side_type_after() != 0 and wplist[wnext].get_side_type_before() != 0) or (witer == 0 and wnext == 0):    
            ipos = swnext_pos
            if ipos > scwplist[scwnext].get_position():
                cwiter = cwnext
                scwiter = scwnext
                cwnext += 1
                scwnext += 1
            while ipos > bnext_pos and bnext+1 != bend:
                last_tangent = curve.derivative(1.0-CUSP_TANGENT_ADJUST)
                biter = bnext
                bnext += 1
                biter_pos = bnext_pos
                bpiter += 1
                hbpiter += 1
                bnext_pos = bline_pos[bpiter]
                hbnext_pos = hbline_pos[hbpiter]
            middle_corner = False
            continue

        if middle_corner == True:
            if split_flag:
                i = scwplist[scwiter]
                n = scwplist[scwnext]
                if not fast_:
                    i = cwplist[cwiter]
                    n = cwplist[cwnext]
                p = ipos
                if not fast_:
                    p = hipos
                add_cusp(side_a,
                        side_b,
                        bline_list[biter].get_vertex(),
                        curve.derivative(CUSP_TANGENT_ADJUST),
                        last_tangent,
                        gv*(expand+width*0.5*widthpoint_interpolate(i, n, p, smoothness)),
                        cusp_type)
            middle_corner = False
            ipos = ipos + EPSILON
        # Secondary loop. For interpolation tasks
        while True:
            swnext_pos = swplist[swnext].get_position()
            if ipos > swnext_pos and bnext_pos >= swnext_pos:
                unitary = None
                ipos = swnext_pos
                hipos = wnext_pos
                q = bline_to_bezier(ipos, biter_pos, bezier_size)
                if q < EPSILON:
                    unitary = iter_t.norm()
                elif q > (1.0 - EPSILON):
                    unitary = next_t.norm()
                else:
                    unitary = curve.derivative(q).norm()
                d = unitary.perp()
                p = curve.value(q)
                ww = 0.0

                if wplist[wnext].get_side_type_before() != 0:
                    ww = 0.0
                else:
                    if wplist[wnext].get_dash():
                        i = scwplist[scwiter]
                        n = scwplist[scwnext]
                        if not fast_:
                            i = cwplist[cwiter]
                            n = cwplist[cwnext]
                        p_my = ipos
                        if not fast_:
                            p_my = hipos
                        wplist[wnext].set_width(widthpoint_interpolate(i, n, p_my, smoothness))
                    ww = wplist[wnext].get_width()
                w = gv*(expand+width*0.5*ww)
                side_a.append([p+d*w, Vector(0, 0), Vector(0, 0)])
                side_b.append([p-d*w, Vector(0, 0), Vector(0, 0)])
                break
            elif ipos > bnext_pos and bnext_pos < swnext_pos:
                hipos = hbnext_pos
                ipos = bnext_pos
                middle_corner = True
                q = bline_to_bezier(ipos, biter_pos, bezier_size)
                q = q if (q > CUSP_TANGENT_ADJUST) else (CUSP_TANGENT_ADJUST)
                q = (1.0 - CUSP_TANGENT_ADJUST) if (q > 1.0 - CUSP_TANGENT_ADJUST) else q
                d = curve.derivative(q).perp().norm()
                p = curve.value(bline_to_bezier(ipos, biter_pos, bezier_size))
                i = scwplist[scwiter]
                n = scwplist[scwnext]
                if not fast_:
                    i = cwplist[cwiter]
                    n = cwplist[cwnext]
                po = ipos
                if not fast_:
                    po = hipos
                w = gv*(expand+width*0.5*widthpoint_interpolate(i, n, po, smoothness))
                side_a.append([p+d*w, Vector(0, 0), Vector(0, 0)])
                side_b.append([p-d*w, Vector(0, 0), Vector(0, 0)])
                biter = bnext
                bnext += 1
                biter_pos = bnext_pos
                bpiter += 1
                hbpiter += 1
                bnext_pos = bline_pos[bpiter]
                hbnext_pos = hbline_pos[hbpiter]
                last_tangent = curve.derivative(1.0-CUSP_TANGENT_ADJUST)
                break

            # Add intepolation
            unitary = None
            q = bline_to_bezier(ipos, biter_pos, bezier_size)
            unitary = curve.derivative(q).norm()
            d = unitary.perp()
            p = curve.value(q)

            if cwplist[cwiter].get_position() == 0.0 and cwplist[cwnext].get_position() != 1.0 and inserted_first:
                cwiter = len(cwplist)
                cwiter -= 1
                if inserted_last:
                    cwiter -= 1
            if cwplist[cwnext].get_position() == 1.0 and inserted_last:
                cwnext = 0
                if inserted_first:
                    cwnext += 1
            i = scwplist[scwiter]
            n = scwplist[scwnext]
            if not fast_:
                i = cwplist[cwiter]
                n = cwplist[cwnext]
            po = ipos
            if not fast_:
                po = std_to_hom(bline_list, ipos, wplistloop, blineloop, fr)
            w = 0.0
            if done_tip:
                w = 0
                done_tip = False
            else:
                w = (gv*(expand+width*0.5*widthpoint_interpolate(i, n, po, smoothness)))
            side_a.append([p+d*w, Vector(0, 0), Vector(0, 0)])
            side_b.append([p-d*w, Vector(0, 0), Vector(0, 0)])
            ipos = ipos + step

    if blineloop:
        side_b.reverse()
        # Remove all data after we encounter a nan value
        # From add_polygon()
        side_a = remove_after_null(side_a)
        side_a.append(side_a[0])    # Closing the contour:: contour->close()
        side_b = remove_after_null(side_b)
        side_b.append(side_b[0])    # Closing the contour:: contour->close()

        side_a.extend(side_b)
        return side_a


    while len(side_b) != 0:
        side_a.append(side_b[-1])
        side_b.pop()
    side_a = remove_after_null(side_a)
    side_a.append(side_a[0])        # Closing the contour::   contour->close()

    return side_a


def remove_after_null(side):
    """
    Modifies the list according to this condition: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/layers/layer_polygon.cpp#L102
    Args:
        side (list[(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector)]): List of points, tangent1, and tangent2

    Returns:
        (list[(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector)]): Modified list
    """
    itr = 0
    while itr < len(side):
        if side[itr][0].isnan():
            break
        itr += 1
    return side[:itr]


def add_cusp(side_a, side_b, vertex, curr, last, w, cusp_type):
    """
    Synfig function: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/modules/mod_geometry/advanced_outline.cpp#L1422

    Args:
        side_a (list[(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector)]): One side of the outline
        side_b (list[(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector)]): Other side of the outline
        vertex (Float): 
        curr (common.Vector.Vector):
        last (common.Vector.Vector):
        w (Float):
        cusp_type (Int): Type of cusp at a frame

    Returns:
        (None)
    """
    CUSP_THRESHOLD = 0.40
    SPIKE_AMOUNT = 4
    SAMPLES = 50
    t1 = last.perp().norm()
    t2 = curr.perp().norm()
    cross = t1*t2.perp()
    perp = (t1-t2).mag()
    if cusp_type == 0:
        if cross > CUSP_THRESHOLD:
            p1 = vertex + t1*w
            p2 = vertex + t2*w
            side_a.append([line_intersection(p1, last, p2, curr), Vector(0, 0), Vector(0, 0)])
        elif cross < -CUSP_THRESHOLD:
            p1 = vertex - t1*w
            p2 = vertex - t2*w
            side_b.append([line_intersection(p1, last, p2, curr), Vector(0, 0), Vector(0, 0)])
        elif cross > 0 and perp > 1:
            amount = max(0.0, (cross/CUSP_THRESHOLD))*(SPIKE_AMOUNT-1)+1
            side_a.append([vertex+(t1+t2).norm()*w*amount, Vector(0, 0), Vector(0, 0)])
        elif cross < 0 and perp > 1:
            amount = max(0.0, (-cross/CUSP_THRESHOLD))*(SPIKE_AMOUNT-1)+1
            side_b.append([vertex-(t1+t2).norm()*w*amount, Vector(0, 0), Vector(0, 0)])
    elif cusp_type == 1:
        if cross > 0:
            p1 = vertex + t1*w
            p2 = vertex + t2*w
            offset = t1.angle()
            angle = t2.angle()-offset
            if angle < RadAngle(0) and offset > RadAngle(0):
                angle += DegAngle(360)
                offset += DegAngle(360)
            tangent = 4 * ((2 * CosAngle(angle/2).get() - CosAngle(angle).get() - 1)/ SinAngle(angle).get())
            curve = Hermite(p1,
                            p2,
                            Vector(-tangent*w*SinAngle(angle*0+offset).get(), tangent*w*CosAngle(angle*0+offset).get()),
                            Vector(-tangent*w*SinAngle(angle*1+offset).get(), tangent*w*CosAngle(angle*1+offset).get()))
            n = 0.0
            while n < 0.999999:
                side_a.append([curve.value(n), Vector(0, 0), Vector(0, 0)]) 
                n += 4.0 / SAMPLES

        if cross < 0:
            p1 = vertex - t1*w
            p2 = vertex - t2*w
            offset = t2.angle()
            angle = t1.angle() - offset
            if angle < RadAngle(0) and offset > RadAngle(0):
                angle += DegAngle(360)
                offset += DegAngle(360)
            tangent = 4 * ((2 * CosAngle(angle/2).get() - CosAngle(angle).get() - 1)/ SinAngle(angle).get())
            curve = Hermite(p1,
                            p2,
                            Vector(-tangent*w*SinAngle(angle*1+offset).get(), tangent*w*CosAngle(angle*1+offset).get()),
                            Vector(-tangent*w*SinAngle(angle*0+offset).get(), tangent*w*CosAngle(angle*0+offset).get()))
            n = 0.0
            while n < 0.999999:
                side_b.append([curve.value(n), Vector(0, 0), Vector(0, 0)]) 
                n += 4.0 / SAMPLES
    return
             

def add_tip(side_a, side_b, vertex, tangent, wp, gv, width, expand):
    """
    Synfig function: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/modules/mod_geometry/advanced_outline.cpp#L1302

    Args:
        side_a (list[(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector)]): One side of the outline
        side_b (list[(common.Vector.Vector, common.Vector.Vector, common.Vector.Vector)]): Other side of the outline
        vertex (Float):
        tangent (common.Vector.Vector):
        wp (common.WidthPoint.WidthPoint):
        gv (Float):
        width (Float): width at a particular frame
        expand (Float): value of expand at a frame

    Returns:
        (None)
    """
    ROUND_END_FACTOR = 4
    SAMPLES = 50
    w = gv * (expand + width*0.5*wp.get_width())
    if wp.get_side_type_before() == 1:
        curve = Hermite(vertex-tangent.perp()*w,
                        vertex+tangent.perp()*w,
                        -tangent*w*ROUND_END_FACTOR,
                        tangent*w*ROUND_END_FACTOR)
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
        n = 0.0
        while n < 0.499999:
            side_a.append([curve.value(0.5+n), Vector(0, 0), Vector(0, 0)]) 
            side_b.append([curve.value(0.5-n), Vector(0, 0), Vector(0, 0)]) 
            n += 2.0/SAMPLES
        side_a.append([curve.value(1.0), Vector(0, 0), Vector(0, 0)])
        side_b.append([curve.value(0.0), Vector(0, 0), Vector(0, 0)])
    elif wp.get_side_type_before() == 2:
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex-tangent*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex+(tangent.perp()-tangent)*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex+tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex-tangent*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex+(-tangent.perp()-tangent)*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex-tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
    elif wp.get_side_type_before() == 3:
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex-tangent*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex+tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex-tangent*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex-tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
    elif wp.get_side_type_before() == 4:
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])

    if wp.get_side_type_after() == 1:
        curve = Hermite(vertex-tangent.perp()*w,
                        vertex+tangent.perp()*w,
                        tangent*w*ROUND_END_FACTOR,
                        -tangent*w*ROUND_END_FACTOR)
        n = 0.0
        while n < 0.499999:
            side_a.append([curve.value(1-n), Vector(0, 0), Vector(0, 0)]) 
            side_b.append([curve.value(n), Vector(0, 0), Vector(0, 0)]) 
            n += 2.0/SAMPLES
        side_a.append([curve.value(0.5), Vector(0, 0), Vector(0, 0)]) 
        side_b.append([curve.value(0.5), Vector(0, 0), Vector(0, 0)]) 
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)]) 
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)]) 
    elif wp.get_side_type_after() == 2:
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex+tangent*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex+(-tangent.perp()+tangent)*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex-tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex+tangent*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex+(tangent.perp()+tangent)*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex+tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
    elif wp.get_side_type_after() == 3:
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex+tangent*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex-tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex+tangent*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex+tangent.perp()*w, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
    elif wp.get_side_type_after() == 4:
        side_a.append([vertex, Vector(0, 0), Vector(0, 0)])
        side_b.append([vertex, Vector(0, 0), Vector(0, 0)])
    return 


def bline_to_bezier(bline_pos, origin, bezier_size):
    """
    Synfig function: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/modules/mod_geometry/advanced_outline.cpp#L1289

    Args:
        bline_pos (Float):
        origin (Float):
        bezier_size (Float):

    Returns:
        (Float)
    """
    if bezier_size != 0:
        return (bline_pos - origin)/bezier_size
    return bline_pos


def hom_to_std(bline_list, pos, index_loop, bline_loop, fr):
    """
    Synfig function: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_bline.cpp#L346

    Args:
        bline_list (list[common.BlinePoint.BlinePoint]): Synfig format bline list
        pos (Float):
        index_loop (Bool):
        bline_loop (Bool):
        fr (Int): Value of frame 

    Returns:
        (Float)
    """
    size = len(bline_list)
    from_vertex = 0

    if pos == 0.0 or pos == 1.0:
        return pos
    if not bline_loop:
        size -= 1
    if size < 1:
        return 0.0
    int_pos = int(pos)
    one = 0.0

    if index_loop:
        pos = pos - int_pos
        if pos < 0:
            pos += 1
            one = 1.0
    else:
        if pos < 0: pos = 0
        if pos > 1: pos = 1

    tl = 0
    pl = 0
    bl = 0
    lengths = []
    tl = bline_length(bline_list, bline_loop, lengths, fr)
    mpl = pos*tl
    nxt = 0
    itr = 0
    if bline_loop:  itr = len(bline_list) - 1
    else:
        itr = nxt
        nxt = nxt + 1
    liter = 0

    while mpl > pl and nxt != len(bline_list):
        pl += lengths[liter]
        bl = lengths[liter]
        itr = nxt
        nxt += 1
        liter += 1
        from_vertex += 1

    if pl > mpl:
        liter -= 1
        nxt -= 1
        if nxt == 0:
            itr = len(bline_list) - 1
        else:
            itr -= 1
        pl -= lengths[liter]
        from_vertex -= 1

    blinepoint0 = bline_list[itr]

    a = blinepoint0.get_vertex()
    c = blinepoint0.get_tangent2()

    if nxt == len(bline_list):
        b = Vector(0, 0)
        d = Vector(0, 0)
    else:
        blinepoint1 = bline_list[nxt]
        b = blinepoint1.get_vertex()
        d = blinepoint1.get_tangent1()

    curve = Hermite(a, b, c, d)

    sn = 0.0
    sn1 = 0.0
    sn2 = 1.0
    t0 = ((mpl - pl)/bl)
    iterations = 0
    max_iterations = 100
    max_error = 0.00001
    error = 0
    fsn1 = t0 - curve.find_distance(0.0, sn1)/bl
    fsn2 = t0 - curve.find_distance(0.0, sn2)/bl
    fsn = 0

    # DO WHILE LOOP IN PYTHON
    while True:
        sn = sn1 - fsn1*((sn1-sn2)/(fsn1-fsn2))
        fsn = t0 - curve.find_distance(0.0, sn)/bl
        sn2 = sn1
        sn1 = sn
        fsn2 = fsn1
        fsn1 = fsn
        error = math.fabs(fsn2 - fsn1)
        iterations += 1
        if not (error > max_error and max_iterations > iterations):
            break

    return int_pos + (from_vertex + sn)/size - one

def std_to_hom(bline_list, pos, index_loop, bline_loop, fr):
    """
    Synfig function: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_bline.cpp#L292

    Args:
        bline_list (list[common.BlinePoint.BlinePoint]): Synfig format bline list
        pos (Float):
        index_loop (Bool):
        bline_loop (Bool):
        fr (Int): Frame number currently processing

    Returns:
        (Float)
    """
    size = len(bline_list)
    if pos == 0.0 or pos == 1.0:
        return pos
    if not bline_loop:
        size -= 1
    if size < 1:
        return 0.0
    int_pos = int(pos)
    one = 0.0

    if index_loop:
        pos = pos - int_pos
        if pos < 0:
            pos += 1
            one = 1.0
    else:
        if pos < 0: pos = 0
        if pos > 1: pos = 1

    tl = 0
    pl = 0
    lengths = []
    tl = bline_length(bline_list, bline_loop, lengths, fr)
    if tl == 0.0:   return pos
    from_vertex = int(pos*size)

    liter = 0
    i = 0
    while i < from_vertex:
        pl += lengths[liter]
        i += 1
        liter += 1

    nxt = 0 # bline.begin()
    itr = 0
    if bline_loop:
        itr = len(bline_list) - 1
    else:
        itr = nxt
        nxt = nxt + 1
    if from_vertex > size - 1:
        from_vertex = size - 1
    bline_point0 = bline_list[nxt+from_vertex-1] if from_vertex else bline_list[itr]
    bline_point1 = bline_list[nxt+from_vertex]
    curve = Hermite(bline_point0.get_vertex(),
                    bline_point1.get_vertex(),
                    bline_point0.get_tangent2(),
                    bline_point1.get_tangent1())
    pl += curve.find_distance(0.0, pos*size - from_vertex)

    return int_pos + pl/tl-one


def bline_length(bline_list, bline_loop, lengths, fr):
    """
    Synfig function: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_bline.cpp#L441

    Args:
        bline (common.Bline.bline): Lottie format bline
        bline_loop (Bool):
        lengths (list[Float]):
        fr (Int): Frame number currently processing

    Returns:
        (Float)
    """
    size = len(bline_list)
    if not bline_loop:  size -= 1
    if size < 1:    return 0.0
    tl = 0

    nxt = 0 # nxt = bline.begin()
    itr = 0
    if bline_loop:
        itr = len(bline_list) - 1
    else:
        itr = nxt
        nxt = nxt + 1
    while nxt != len(bline_list):
        blinepoint0 = bline_list[itr]
        blinepoint1 = bline_list[nxt]
        curve = Hermite(blinepoint0.get_vertex(),
                        blinepoint1.get_vertex(),
                        blinepoint0.get_tangent2(),
                        blinepoint1.get_tangent1())
        l = curve.length()
        if lengths is not None:
            lengths.append(l)
        tl += l
        itr = nxt
        nxt += 1
    return tl

def widthpoint_interpolate(prev, nxt, p, smoothness=0.0):
    """
    Synfig function: https://github.com/synfig/synfig/blob/15607089680af560ad031465d31878425af927eb/synfig-core/src/synfig/valuenodes/valuenode_wplist.cpp#L111

    Args:
        prev (common.WidthPoint.WidthPoint)
        nxt (common.WidthPoint.WidthPoint)
        p (Float)
        smoothness (Float)

    Returns:
        (Float)
    """
    side_int = 0
    #nw, pw = 0, 0
    rw = 0.0
    epsilon = 0.0000001
    np = nxt.get_position()
    pp = prev.get_position()
    nw = nxt.get_width()
    pw = prev.get_width()
    nsb = nxt.get_side_type_before()
    nsa = nxt.get_side_type_after()
    psb = prev.get_side_type_before()
    psa = prev.get_side_type_after()

    if p == np:
        return nw
    if p == pp:
        return pw
    if np > pp:
        if np > p and p > pp:
            q = 0
            if nsb != side_int:
                nw = 0.0
            if psa != side_int:
                pw = 0.0
            if (np - pp < epsilon):
                q = 0.5
            else:
                q = (p - pp) / (np - pp)
            rw = pw + (nw-pw)*(q*(1.0-smoothness)+q*q*q*(10+q*(6*q-15))*smoothness)
        elif p < pp:
            if psb != side_int:
                pw = 0.0
            rw = pw
        elif p > np:
            if nsa != side_int:
                nw = 0.0
            rw = nw
    elif p > pp or np > p:
        q = 0.0
        if nsb != side_int:
            nw = 0.0
        if psa != side_int:
            pw = 0.0
        if np+1.0-pp < epsilon:
            q = 0.5
        else:
            if p > pp:
                q = (p - pp)/(np+1.0-pp)
            if np > p:
                q = (p+1.0-pp)/(np+1.0-pp)
        rw = pw+(nw-pw)*(q*(1.0-smoothness)+q*q*q*(10+q*(6*q-15))*smoothness)
    elif p > np and p < pp:
        #q = 0.0
        if nsa != side_int:
            nw = 0.0
        if psb != side_int:
            pw = 0.0
        if pp - np < epsilon:
            q = 0.5
        else:
            q = (p - np) / (pp - np)
        rw = nw+(pw-nw)*(q*(1.0-smoothness)+q*q*q*(10+q*(6*q-15))*smoothness) 
    return rw
